/********************************************************************/

#include "Tool.h"

/********************************************************************/

// Prints out what the add options for filename case and space conversions
// are. This is used when --no-action/-n is in effect
static void printAddOptions (CommandStruct *command)
{
    printf ("    -> ");

    if (command->keepIncomingCase)
        printf ("Case of added filenames will be preserved, ");
    else
        printf ("Added filenames will be converted to lowercase, ");

    printf ("spaces in filenames will be ");

    switch (command->spaceAction)
    {
        case SPACES_KEEP: 
            printf ("preserved, ");
            break;

        case SPACES_CONVERT:
            printf ("converted to underscores, ");
            break;

        case SPACES_REMOVE:
            printf ("removed, ");
            break;
    }

    printf ("existing files ");

    if (command->clobberIncoming)
        printf ("WILL be clobbered by new files\n");
    else
        printf ("WILL NOT be clobbered by new files\n");
}

/********************************************************************/

// This takes an add command and configuration information. If the
// --no-action/-n option is in effect, this will output text to display
// what the command would do.
static void printAddCommand (CommandStruct *command , ConfigInfo *config)
{
    // If we are not just testing, do nothing here.
    if (config->isTesting == SDL_FALSE)
        return;

    // Based on what add command this is, act accordingly.
    switch (command->cmd)
    {
        case CMD_ADD:
            printf ("-> %s: Files in local path '%s' whose names match '%s' will be added to NFI at path '%s'\n" ,
               optionNameFromValue (command->cmd) , command->fromPath , command->arg , command->toPath);
            break;

        case CMD_ADDRECURSE:
            printf ("-> %s: Files in local path '%s' and it's children whose names match '%s' will be added to NFI starting at path '%s'\n" ,
               optionNameFromValue (command->cmd) , command->fromPath , command->arg , command->toPath);
            break;

        case CMD_ADDFILES:
            printf ("-> %s: Files from the command line will be added to NFI starting at path '%s'\n" ,
               optionNameFromValue (command->cmd) , command->toPath);
            break;

        case CMD_ADDDIR:
            printf ("-> %s: All files in '%s' will be added to NFI at path '%s'\n" ,
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        case CMD_ADDDIRRECURSE:
            printf ("-> %s: All files and subdirectories of '%s' will be added to NFI starting at path '%s'\n" ,
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        default:
            break;
    }

    // If this is an option that takes command line files, display them now.
    if (command->cmd == CMD_ADDFILES)
    {
        printf ("    -> ");

        for (int i = 0 ; i < config->cmdlineFileCount ; i++)
            printf ("%s " , config->cmdlineFiles[i]);

        printf ("\n");
    }

    // Now print the options to the add command that are current for this command
    printAddOptions (command);
}

/********************************************************************/

// Take the name of a file and what path it should be stored in
// internally, and fills out the given NFI info struct. The name
// will be copied to the name (possibly modified by the config given),
// and the path will be written to the path portion. Everything else
// is left as passed in.
//
// The config given determines if the file is modified on it's way into
// the path. In particular, it might have spaces compressed out or
// replaced with underscores, and the filename might be forced to all
// lowercase.
static void makeNFIName (const char *name , const char *path , NFIParsedDirEntry *info , CommandStruct *config)
{
    // Store the name and path as given.
    strcpy (info->path , path);
    strcpy (info->name , name);

    // Check the config to see if we should lowercase the name of the
    // file or not.
    if (config->keepIncomingCase == SDL_FALSE)
    {
        for (int i = 0 ; info->name[i] != '\0' ; i++)
            info->name[i] = tolower (info->name[i]);
    }

    int i,j;

    // What should be do with spaces in the filename portion?
    switch (config->spaceAction)
    {
        // Do nothing.
        case SPACES_KEEP:
            break;

        // Replace space characters with an underscore.
        case SPACES_CONVERT:
            for (i = 0 ; info->name[i] != '\0' ; i++)
            {
                if (info->name[i] == ' ')
                    info->name[i] = '_';
            }
            break;
            
        // Cut spaces out
        case SPACES_REMOVE:
            // Copy the string to itself, skipping spaces.
            for (i = 0,j = 0; info->name[i] != '\0' ; i++)
            {
                // if this isn't a space, copy it.
                if (info->name[i] != ' ')
                    info->name[j++] = info->name[i];
            }
            info->name[j] = '\0';
            break;
    }
}

/********************************************************************/

// --add: Argument is a regex. Scan the local directory --from,
// all filenames that match the regex are added to the NFI archive
// at the internal path --to
//
// --add-recurse: Argument is a regex. Scan the local directory --from,
// all filenames that match the regex are added to the NFI archive
// at the internal path --to. We also recursively scan files in directories
// contained in --from. In this case, the relative path of the files found
// is added to the --to path to come up with the final destination in 
// the file.
//
// --add-files: Take all files specified on the command line and add them
// to the archive starting at the --to path. The path of the files given is
// added to the --to path to come up with the final destination. If a file
// given on the command line is a directory, it will be recursed into.
//
// --add-dir: Add all files underneath the --from path
// to the NFI archive at the --to path. 
//
// --add-dir-recurse: Add all files and directories underneath the --from path
// to the NFI archive starting at the --to path. The relative path of
// recursed files is added to the --to path to come up with the final destination
//
// This function performs all of the above add commands, since they all do the
// same things, with only slight differences in how they select files for inclusion.
void cmdAdd (NFIFile *nfi , CommandStruct *command , ConfigInfo *config)
{
    // Print out the command, in case we're in no-action mode. This does nothing
    // if we are not.
    printAddCommand (command , config);

    // Pretend that we won't recurse with this command.
    SDL_bool recurse = SDL_FALSE;

    // Check to see if the command we're executing is one of the ones that
    // optionally recurse, and set the recurse flag to true if it is.
    if (command->cmd == CMD_ADDRECURSE || command->cmd == CMD_ADDDIRRECURSE)
        recurse = SDL_TRUE;

    // Create a list of files to scan.  How we do this depends on what command we are. 
    // The --add-files command takes files from the command line, but all the others 
    // scan the from path (possibly recursively).
    ScannedFiles *filesToAdd;

    if (command->cmd == CMD_ADDFILES)
        filesToAdd = new ScannedFiles ((const char **) config->cmdlineFiles , config->cmdlineFileCount , config->isVerbose);
    else
        filesToAdd = new ScannedFiles (command->fromPath , &command->compiledRegex , config->isVerbose , recurse);

    // If we found files, add them.
    if (filesToAdd->filesFound () != 0)
    {
        // Store the current size of the directory in the file, as this will be the index of
        // the first added file.
        int index = nfi->dirHeader.dirSize;
        int addCount=0;
        int i;

        // Reallocate the directory to have the files it had plus the new ones. Also update
        // the dir size.
        nfi->dirHeader.dirSize += filesToAdd->filesFound ();
        nfi->directory = (NFIParsedDirEntry *) realloc (nfi->directory , nfi->dirHeader.dirSize * sizeof (NFIParsedDirEntry));

        // in the NFI path structure list (nfi->pathStruct) we maintain a list of all files being stored in
        // the NFI, along with a pointer to their directory structure in nfi->directory. Because we just
        // reallocated that, the location of existing members may have shifted in memory. This leaves the
        // entry in the path structure wrong.
        //
        // Fortunately, each directory entry keeps a pointer to the PathEntry structure that is referencing
        // it. So we can scan over all directory entries and update entry pointers.
        for (i = 0 ; i < index ; i++)
        {
            // If we have a valid entryPtr, then tell it where we are now located.
            if (nfi->directory[i].entryPtr != NULL)
                nfi->directory[i].entryPtr->info = &nfi->directory[i];
        }

        // Now, add in all of the new files.
        for (int i = 0 ; i < filesToAdd->filesFound () ; i++, index++)
        {
            // First of all, initialize this info struct.
            memset (&nfi->directory[index] , 0 , sizeof (NFIParsedDirEntry));

            // Fill out the name and path. This will fix the case and spaces in the incoming filename as appropriate.
            makeNFIName (filesToAdd->fileAtIndex (i) , command->toPath , &nfi->directory[index] , command);
            
            // Tell the info struct that it is being added
            nfi->directory[index].type = NFI_ADDED_FILE;

            // Store the source file being added
            sprintf (nfi->directory[index].sourceName , "%s/%s" , filesToAdd->scannedDir () , filesToAdd->fileAtIndex (i));

            // Now add the file to the structure. If this works, output a message.
            if (nfi->pathStruct->addFile (&nfi->directory[index] , command->clobberIncoming))
            {
                // Count this as a file added
                addCount++;
            }
            else
            {
                // The add failed, so flag the entry we just added to the structure as a removed file, so that
                // nothing happens. If we don't do that, we will end up adding the file anyway when we write it
                // out. This won't flag as a change because nothing was really done. 
                nfi->directory[index].type = NFI_REMOVED_FILE;
            }
        }

        // If any files were added in the above, then tell the nfi file that, so that it knows to save
        // the changes. Otherwise, nothing needs to be done for this command.
        if (addCount)
            nfi->changeCount++;

        // Get rid of the list of added files
    } else
    {
        // No files were found to add. If we are verbose or testing, say that.
        if (config->isTesting || config->isVerbose)
            printf ("No files found to add\n");
    }

    // Now, get rid of the files we scanned
    delete filesToAdd;

    // All done. If we are outputting what we're doing, seperate this command from any that might follow
    if (config->isTesting)
        printf ("\n");
}

/********************************************************************/

