/********************************************************************/

#include "Tool.h"

/********************************************************************/

// Tells you if a command requires a regex or not
static SDL_bool commandRequiresRegex (CommandType command)
{
    switch (command)
    {
        // These need a regex
        case CMD_ADD:
        case CMD_ADDRECURSE:
        case CMD_REMOVE:
        case CMD_REMOVERECURSE:
        case CMD_EXTRACT:
        case CMD_EXTRACTRECURSE:
            return SDL_TRUE;

        // The rest don't
        default:
            return SDL_FALSE;
    }

    // This keeps compilers happy.
    return SDL_FALSE;
}

/********************************************************************/

// Tells you if a command requires a "to" path to work properly.
static SDL_bool commandRequiresToPath (CommandType command)
{
    switch (command)
    {
        // These need a to
        case CMD_ADD:
        case CMD_ADDRECURSE:
        case CMD_ADDFILES:
        case CMD_ADDDIR:
        case CMD_ADDDIRRECURSE:
        case CMD_MOVE:
        case CMD_MOVERECURSE:
        case CMD_MOVEFILES:
        case CMD_MOVEDIR:
        case CMD_MOVEDIRRECURSE:
        case CMD_EXTRACTFILES:
        case CMD_EXTRACTDIR:
        case CMD_EXTRACTDIRRECURSE:
        case CMD_EXTRACT:
        case CMD_EXTRACTRECURSE:
            return SDL_TRUE;

        // The rest don't.
        default:
            return SDL_FALSE;
    }

    // This keeps compilers happy.
    return SDL_FALSE;
}

/********************************************************************/

// Tells you if a command requires a "from" path to work properly.
static SDL_bool commandRequiresFromPath (CommandType command)
{
    switch (command)
    {
        // These need a from
        case CMD_ADD:
        case CMD_ADDRECURSE:
        case CMD_ADDDIR:
        case CMD_ADDDIRRECURSE:
        case CMD_REMOVE:
        case CMD_REMOVERECURSE:
        case CMD_REMOVEFILES:
        case CMD_REMOVEDIR:
        case CMD_REMOVEDIRRECURSE:
        case CMD_MOVE:
        case CMD_MOVERECURSE:
        case CMD_MOVEFILES:
        case CMD_MOVEDIR:
        case CMD_MOVEDIRRECURSE:
        case CMD_EXTRACTFILES:
        case CMD_EXTRACTDIR:
        case CMD_EXTRACTDIRRECURSE:
        case CMD_EXTRACT:
        case CMD_EXTRACTRECURSE:
            return SDL_TRUE;

        // Everything else doesn't.
        default:
            return SDL_FALSE;
    }

    // This keeps compilers happy.
    return SDL_FALSE;
}

/********************************************************************/

// Tells you if a command requires files provided on the command line
// in order to work properly.
static SDL_bool commandRequiresCmdlineFiles (CommandType command)
{
    switch (command)
    {
        // These need cmdline files
        case CMD_ADDFILES:
        case CMD_REMOVEFILES:
        case CMD_MOVEFILES:
        case CMD_EXTRACTFILES:
            return SDL_TRUE;

        // Everything else doesn't.
        default:
            return SDL_FALSE;
    }

    // This keeps compilers happy
    return SDL_FALSE;
}

/********************************************************************/

// Validate the given command, and return back true or false to indicate
// if the command is valid or not.
//
// This only checks that a command has sensible values. It doesn't check,
// for example, if we're going to try to delete something from a file that
// doesn't exist and so on.
SDL_bool validateCommand (CommandStruct *command , int index , int cmdlineCount)
{
    // Check if the command needs a from path
    if (commandRequiresFromPath (command->cmd) && command->fromPath == NULL)
    {
        fprintf (stderr , "Error in command %d: %s requires a --from path, but one hasn't been set yet!\n\n" , index , optionNameFromValue (command->cmd));
        return SDL_FALSE;
    }

    // Check if the command needs a to path
    if (commandRequiresToPath (command->cmd) && command->toPath == NULL)
    {
        fprintf (stderr , "Error in command %d: %s requires a --to path, but one hasn't been set yet!\n\n" , index , optionNameFromValue (command->cmd));
        return SDL_FALSE;
    }

    // Check if the command needs commandline files.
    if (commandRequiresCmdlineFiles (command->cmd) && cmdlineCount == 0)
    {
        fprintf (stderr , "Error in command %d: %s requires files on the command line, but none were given!\n\n" , index , optionNameFromValue (command->cmd));
        return SDL_FALSE;
    }

    // Check if the command requires a regex.
    if (commandRequiresRegex (command->cmd))
    {
        // The regex is in the command's argument. Try to compile it.
        int result = regcomp (&command->compiledRegex , command->arg , REG_EXTENDED | REG_NOSUB);

        // If it worked, flag the command so that it knows it has a compiled regex to clean up.
        // otherwise, generate an error and return false.
        if (result == 0)
            command->haveRegex = SDL_TRUE;
        else
        {
            char message[256];

            // Determine what the cause of failure was.
            regerror (result , &command->compiledRegex , message , sizeof (message) - 1);

            // Print it and return false.
            fprintf (stderr , "Error in command %d: Bad regex (%s)\n\n" , index , message);
            return SDL_FALSE;
        }
    }

    // It's good
    return SDL_TRUE;
}

/********************************************************************/

// Given an NFIFile and configuration information, this will execute all of the
// commands one by one. This presumes that the commands entered in config have
// been checked to ensure that all needed parameters are provided.
void performCommands (NFIFile *sourceFile , ConfigInfo *config)
{
    int moveCommands = 0;

    // Loop over all commands.
    for (int i = 0 ; i < config->commandCount ; i++)
    {
        // Alias this command
        CommandStruct *cmd = &config->commands[i];

        // Do something based on this command
        switch (cmd->cmd)
        {
            // Add commands
            case CMD_ADD:
            case CMD_ADDRECURSE:
            case CMD_ADDFILES:
            case CMD_ADDDIR:
            case CMD_ADDDIRRECURSE:
                cmdAdd (sourceFile , cmd , config);
                break;

            // Remove commands except --remove-files, which works
            // in a different way
            case CMD_REMOVE:
            case CMD_REMOVERECURSE:
            case CMD_REMOVEDIR:
            case CMD_REMOVEDIRRECURSE:
                cmdRemove (sourceFile , cmd , config);
                break;

            // --remove-files
            case CMD_REMOVEFILES:
                cmdRemoveFiles (sourceFile , cmd , config);
                break;

            // Move commands except --move-files, which works in 
            // a different way
            case CMD_MOVE:
            case CMD_MOVERECURSE:
            case CMD_MOVEDIR:
            case CMD_MOVEDIRRECURSE:
                moveCommands++;
                cmdMove(sourceFile , cmd , config);
                break;

            // --move-files
            case CMD_MOVEFILES:
                moveCommands++;
                cmdMoveFiles (sourceFile , cmd , config);
                break;

            // --extract-files
            case CMD_EXTRACTFILES:
                cmdExtractFiles  (sourceFile , cmd , config);
                break;

            // Extract commands except --extract-files, which works in
            // a different way
            case CMD_EXTRACTDIR:
            case CMD_EXTRACTDIRRECURSE:
            case CMD_EXTRACT:
            case CMD_EXTRACTRECURSE:
                cmdExtract(sourceFile , cmd , config);
                break;

            default:
                fprintf (stderr , "ERROR: Unknown command value %d?!\n" , cmd->cmd);
        }
    }

    // If there were commands executed, and all of them were move commands, then
    // we can tell the NFIFile that. When we close up a modified file and all
    // of the commands were moves, we can do a "quick save" by just rewriting the
    // directory to contain the new filename information, since none of the files
    // have actually changed. This can be a dramatic speedup for large archives.
    if (config->commandCount != 0 && moveCommands == config->commandCount)
        sourceFile->movesOnly = SDL_TRUE;
}

/********************************************************************/

