/********************************************************************/

#include "Tool.h"

/********************************************************************/

// These define the long options that we support. 
// The members are:
//   o Name of the long option
//   o Do we want an argument? 
//   o flag var
//   o flag value
//
// When we call getopt_long and it matches up a long option, it will do 
// one of two things:
//  1) If flag var is a NULL pointer, then getopt_long will
//     return "flag value" so that we know what argument it
//     just hit on. 
//  2) if flag var is non-NULL, then getopt_long will store 
//     the value in flag value into the area pointed at by
//     flag var. In this case, getopt_long returns 0.
//
// You can specify the value of a long argument on the command
// line as:
// 
//     --argument value
//     --argument=value
//
// Note that if a value is optional, you can only set it by
// using the second form above. 
static struct option long_options[] = 
{
    // This specifies the name of the NFI file that we will be
    // working with. This option can only appear once.
    {"file"                    , required_argument, NULL,  CMD_FILE}, // short form: f

    // It is only valid to specify one of these three parameters.
    // The first says that we will create a new archive (and fail
    // if it exists), the second says that we will be updating an
    // existing archive (fail if it doesn't exist), and the last 
    // says that we will be listing the contents of the archive
    // given by file (fail if it doesn't exist).
    {"create"                  , no_argument,       NULL,  CMD_CREATE}, // short form: c
    {"update"                  , no_argument,       NULL,  CMD_UPDATE}, // short form: u
    {"list"                    , no_argument,       NULL,  CMD_LIST}, // short form: l

    // This makes the tool be more verbose in what it's doing. Useful to watch
    // what is happening if an error occurs, etc.
    {"verbose"                 , no_argument,       NULL,  CMD_VERBOSE}, // short form: v

    // Turns on debug mode, which is sort of like "super verbose", in that output to 
    // help in debugging is output. This works independantly of the verbose setting,
    // this only outputs data that may be useful while debugging a problem.
    {"debug"                   , no_argument,       NULL,  CMD_DEBUG}, // short form: d

    // This flag indicates that the tool should gather up the arguments as if
    // it was going to run, then say what it would have done had it run, but
    // not actually do anything. Useful for debugging a command
    {"no-action"               , no_argument,       NULL,  CMD_TESTING}, // short form: n

    // Display version information and quit.
    {"version"                 , no_argument,       NULL,  CMD_VERSION}, // short form: V

    // Display help for the command line and then quit.
    {"help"                    , no_argument,       NULL,  CMD_HELP}, // short form; H

    // This sets the directory that the tool will use when it needs to create
    // temporary files. The default is one of: $TMPDIR, $TMP, /tmp (in that order).
    {"tmpdir"                  , required_argument, NULL,  CMD_TMPDIR},

    // These two options control what happens to the case of files that are added
    // to an NFI archive. We can either keep the existing case or convert the
    // filename to all lowercase (default).
    {"keep-case"               , no_argument,       NULL,  CMD_KEEPCASE},
    {"lowercase"               , no_argument,       NULL,  CMD_LOWCASE},

    // These three options control what happens to spaces in files that are added to
    // an NFI archive. We can either keep them (default), convert them to underscores,
    // or remove them altogether.
    {"keep-spaces"             , no_argument,       NULL,  CMD_KEEPSPACE},
    {"convert-spaces"          , no_argument,       NULL,  CMD_CONVSPACE},
    {"remove-spaces"           , no_argument,       NULL,  CMD_DELSPACE},
    {"clobber"                 , no_argument,       NULL,  CMD_CLOBBER},
    {"noclobber"               , no_argument,       NULL,  CMD_NOCLOBBER},

    // These options control the output of the --list/-l command. 
    //
    // The sort command takes an option argument that is one of name,size,offset or path,
    // or a single character n,s,o,p. This option assumes you passed in "name" if no
    // option is specified. This controls the sort order of files when they are listed.
    //
    // If you specify human/-h, the sizes of files will be display as K, M or G, as 
    // appropriate, or as just bytes if < 1024.
    //
    // descending reverses the normal sort order for sorts.
    //
    // no-size turns off the output of size information when a brief display is being
    // used. It has no effect otherwise. This is useful for shell scripts and such
    // which may want the contents of the file without having to do complex parsing.
    {"sort"                    , optional_argument, NULL,  CMD_SORT},
    {"descending"              , no_argument,       NULL,  CMD_DESCEND},
    {"brief"                   , no_argument,       NULL,  CMD_BRIEF},
    {"no-size"                 , no_argument,       NULL,  CMD_NOSIZE},
    {"human"                   , no_argument,       NULL,  CMD_HUMAN}, // short form: h

    // These two options specify a from and to argument (usually a path spec inside an
    // NFI file, but not neccesarily that) that is used by the add/remove/move/extract
    // options.
    {"from"                    , required_argument, NULL,  CMD_FROM},
    {"to"                      , required_argument, NULL,  CMD_TO},

    // These options are used to add files:
    //
    // add and add-recurse take a regex argument and scan for files that match the regex.
    // Any that match are added into the NFI archive at the path specified by --to
    //
    // add-files adds all files that were specified on the command line to the --to directory.
    //
    // add-dir and add-dir-recurse scan the directory in --from and add all the files found there
    // to the --to directory inside the NFI archive.
    {"add"                     , required_argument, NULL,  CMD_ADD},
    {"add-recurse"             , required_argument, NULL,  CMD_ADDRECURSE},
    {"add-files"               , no_argument,       NULL,  CMD_ADDFILES},
    {"add-dir"                 , no_argument,       NULL,  CMD_ADDDIR},
    {"add-dir-recurse"         , no_argument,       NULL,  CMD_ADDDIRRECURSE},

    // These options are used to remove files:
    //
    // remove and remove-recurse take a regex and remove all files in the --from path in the NFI
    // if they match the regex. The regex only matches on the filename portion. 
    //
    // remove-files removes all files given on the command line from the archive if they exist in the
    // --from path in the NFI.
    //
    // remove-dir and remove-dir-recurse remove the contents of entire directory/subdirectory trees
    // as specified by the --from parameter.
    {"remove"                  , required_argument, NULL,  CMD_REMOVE},
    {"remove-recurse"          , required_argument, NULL,  CMD_REMOVERECURSE},
    {"remove-files"            , no_argument,       NULL,  CMD_REMOVEFILES},
    {"remove-dir"              , no_argument,       NULL,  CMD_REMOVEDIR},
    {"remove-dir-recurse"      , no_argument,       NULL,  CMD_REMOVEDIRRECURSE},

    // These options are used to move files:
    //
    // move and move-recurse take a regex and move all files whose names match
    // the regex and are contained in the --from path to the --to path.
    //
    // move-files moves all files given on the command line that
    // exist in the --from path in the NFI to the --to path in the NFI.
    // 
    // move-dir and move-dir-recurse move all files in the --from 
    // path in the NFI archive to the --to path in the NFI archive.
    {"move"                    , required_argument, NULL,  CMD_MOVE},
    {"move-recurse"            , required_argument, NULL,  CMD_MOVERECURSE},
    {"move-files"              , no_argument,       NULL,  CMD_MOVEFILES},
    {"move-dir"                , no_argument,       NULL,  CMD_MOVEDIR},
    {"move-dir-recurse"        , no_argument,       NULL,  CMD_MOVEDIRRECURSE},

    // These options are used to extract files:
    //
    // extract-files extracs all files given on the command line that exist
    // in the --from path in the NFI to the --to path in the real filesystem.
    //
    // extract-dir and extract-dir-recurse extract all files in the --from
    // NFI path to the local filesystem --to path.
    //
    // extract-regex and extract-regex-recurse takes a regex. All files 
    // in the --from NFI path are compared to the given regex (filename portion only)
    // and all that match are extracted to the --to real filesystem path.
    {"extract-dir"             , no_argument,       NULL,  CMD_EXTRACTDIR},
    {"extract-dir-recurse"     , no_argument,       NULL,  CMD_EXTRACTDIRRECURSE},
    {"extract-files"           , no_argument,       NULL,  CMD_EXTRACTFILES},
    {"extract"           , required_argument,       NULL,  CMD_EXTRACT},
    {"extract-recurse"   , required_argument,       NULL,  CMD_EXTRACTRECURSE},

    // This is the command seperator. When this is encountered, a command is built from the
    // current state of things (as set by previous command line arguments) and added to the
    // command list. There is an implicit --then command as the last item in the command line,
    // so that simple commands do not need you to add this yourself, though you can (this would
    // add the same command twice, however).
    {"then"                  , no_argument,       NULL,  CMD_THEN},

    // Terminates the list
    {0 , 0, NULL,  0}
};

/********************************************************************/

// Prints out help on how to use the tool to stdout.
void printCommandHelp (void)
{
    printf (
        "Usage: nfitool [-cul] [-nvVHh] --file/-f archive [--tmpdir path] [command ...]  [file ...]\n \n"
        "Specify one of c(create) u(update) or l(list), plus the name of an\n archive file.\n \n"
        "--tmpdir specifies where temporary files are written to. Default is $TMPDIR,\n"
        "$TMP or /tmp.\n \n"
        "General options\n"
        "  --no-action : Say what would be done, but do not save a new archive.\n"
        "  --verbose   : Be verbose about what is going on\n"
        "  --debug     : Display debugging information\n"
        "  --version   : Display version and exit\n"
        "  --help      : Display help and exit\n"
        "  --to        : Set destination path for commands\n"
        "  --from      : Set source path for commands\n"
        "  --then      : Seperate multiple commands\n \n"
        "Adding Files\n"
        "  --add-files, --add=regex, --add-recurse=regex,\n"
        "  --add-dir,--add-dir-recurse\n \n"
        "Removing Files\n"
        "  --remove-files, --remove=regex, --remove-recurse=regex,\n"
        "  --remove-dir,--remove-dir-recurse\n \n"
        "Moving Files\n"
        "  --move-files, --move=regex, --move-recurse=regex,\n"
        "  --move-dir, --move-dir-recurse\n \n"
        "Extracting files\n"
        "  --extract-files, --extract-regex=regex, --extract-regex-recurse=regex,\n"
        "  --extract-dir, --extract-dir-recurse\n \n \n"
        "See the man page nfitool(1) for more information.\n \n"
    );
}

/********************************************************************/

// Convert an integer that represnets a command into a string suitable
// for display purposes.
//
// If value is one of the CMD_* values, the appropriate long option is
// returned. If it is a character constant, then the appropriate short
// option is returned.
char *optionNameFromValue (int value)
{
    static char buffer[50];

    // If the value isn't one of our special codes, it must be a short
    // option, so turn it into a string and return it.
    if (value < CMD_FILE)
    {
        sprintf (buffer , "-%c" , value);
        return buffer;
    }

    // Loop over all long options
    for (int i = 0 ; long_options[i].name != NULL ; i++)
    {
        // If this is it return it's name.
        if (long_options[i].val == value)
        {
            sprintf (buffer , "--%s" , long_options[i].name);
            return buffer;
        }
    }

    // We don't know what it is
    return "--what-the-hell";
}

/********************************************************************/

// Checks to see if TMPDIR is set, if so, return that. If not, then
// if TMP is set, return that. if not, return /tmp. The string returned
// is allocated via strdup and should be freed by the caller.
static char *getDefaultTempDir (void)
{
    // Get the value of TMPDIR
    char *temp = getenv ("TMPDIR");

    // If that didn't work or it's empty, try again
    if (temp == NULL || *temp == '\0')
    {
        temp = getenv ("TMP");

        // If that didn't work, then set it to the default.
        if (temp == NULL || *temp == '\0')
            temp = "/tmp";
    }

    // Return a duplicate of what we stopped on.
    return strdup (temp);
}

/********************************************************************/

// Validate to make sure that the config information that we created
// is valid.
static SDL_bool validateConfig (ConfigInfo *config)
{
    // The user should have specified 1 of create/update/list. Check
    // to see if any options were set.
    int counter=0;
    if (config->isCreating) counter++;
    if (config->isUpdating) counter++;
    if (config->isListing) counter++;

    // If counter is 0, then none were specified, in which case, turn on
    // the "show help" option and return success. This will force us to
    // display help and exit. 
    if (counter == 0)
    {
        config->showHelp = SDL_TRUE;
        return SDL_TRUE;
    }

    // Otherwise, there should only be 1 thing
    if (counter != 1)
    {
        fprintf (stderr , "You must specify only one of create, update or list!\n\n");
        return SDL_FALSE;
    }

    // If there is no filename to load, we're unhappy.
    if (config->nfiFilename == NULL)
    {
        fprintf (stderr , "You must specify an NFI file to work with!\n\n");
        return SDL_FALSE;
    }

    // If there is no temporary directory, set one up.
    if (config->tmpDir == NULL)
        config->tmpDir = getDefaultTempDir ();

    // Validate all commands
    if (config->commandCount != 0)
    {
        // Loop over all commands. If any fail validation, leave. The validator
        // says what's wrong so we don't have to. The index we pass in is used for
        // display purposes so we add 1 to it to make it human readable.
        for (int i = 0 ; i < config->commandCount ; i++)
        {
            if (validateCommand (&config->commands[i] , i + 1 , config->cmdlineFileCount) == SDL_FALSE)
                return SDL_FALSE;
        }
    }

    // Must have worked!
    return SDL_TRUE;
}

/********************************************************************/

// Close up and release all memory used by the config information
// passed in. This works even if info is NULL or only partially
// filled out
void closeConfigInfo (ConfigInfo *info)
{
    int i;

    // Leave if nothing
    if (info == NULL)
        return;

    // Clear out the names if there are any
    if (info->nfiFilename != NULL) free (info->nfiFilename);
    if (info->tmpDir != NULL) free (info->tmpDir);
    if (info->fromPath != NULL) free (info->fromPath);
    if (info->toPath != NULL) free (info->toPath);

    // if there is a command arg, toss it as well.
    if (info->currentCommandArg != NULL) free (info->currentCommandArg);

    // Release the command table. We will alias the table to make
    // this part easier.
    CommandStruct *table = info->commands;
    for (i = 0 ; i < info->commandCount ; i++)
    {
        if (table[i].arg != NULL)      free (table[i].arg);
        if (table[i].toPath != NULL)   free (table[i].toPath);
        if (table[i].fromPath != NULL) free (table[i].fromPath);
        if (table[i].haveRegex)        regfree (&table[i].compiledRegex);
    }

    // Get rid of any command line files we may have stored.
    if (info->cmdlineFiles != NULL)
    {
        // Release all stored filenames
        for (i = 0 ; i < info->cmdlineFileCount ; i++)
            free (info->cmdlineFiles[i]);

        // Now free the array
        free (info->cmdlineFiles);
    }

    // Now get rid of the table itself.
    free (table);

    // Free the config structure now
    free (info);
}

/********************************************************************/

// Takes a string and determines what sort of sort is required. If
// style is NULL, the sort style is name. If the style is unknown,
// the sort style is natural.
static SortStyle getSortStyle (char *style)
{
    // No style defaults to name.
    if (style == NULL)
        return SORT_NAME;

    if (strcmp (style , "name") == 0 || strcmp (style , "n") == 0)
        return SORT_NAME;

    if (strcmp (style , "size") == 0 || strcmp (style , "s") == 0)
        return SORT_SIZE;

    if (strcmp (style , "offset") == 0 || strcmp (style , "o") == 0)
        return SORT_OFFSET;

    if (strcmp (style , "path") == 0 || strcmp (style , "p") == 0)
        return SORT_PATH;

    // Couldn't determine the sort order.
    return SORT_NATURAL;
}

/********************************************************************/

// Take a path that is part of an NFI path, and make sure that it has
// a leading and trailing seperator character, and that each component
// is seperated by exactly one path separator. The return value is either
// a strdup'd copy of the input path all fixed up, or NULL if the path
// passed in was also NULL. 
static char *fixCommandPath (const char *path)
{
    char buffer[NFI_NAMELEN + 1]="/";
    char copy[NFI_NAMELEN + 1];
    char *part;
    int pos=0;

    // if we were given no path, leave now.
    if (path == NULL)
        return NULL;

    // Make a copy of the path passed in.
    strcpy (copy , path);

    // Keep pulling path components out of of
    // the path. Each time we get one, append
    // it and a trailing "/" character to the
    // new path (which starts out holding just
    // a single path seperator. This operation
    // clobbers the toPath in the command, but
    // that is ok because we will be releasing
    // it anyway.
    part = nextPathPart (copy , &pos);
    while (part != NULL)
    {
        strcat (buffer , part);
        strcat (buffer , "/");
        part = nextPathPart (copy , &pos);
    }

    // Now, return a duplicate copy of our buffer.
    return strdup (buffer);
}

/********************************************************************/

// This is called in response to the --then command, and once after
// all arguments have been parsed out, to use the stored information
// about command, it's argument, and so on into a command. 
//
// The created command is added to the config info provided.
static void addCommandToConfig (ConfigInfo *config)
{
    // If there is not enough room to add in the new command,
    // enlarge the table.
    if (config->commandCount == config->commandTableSize)
    {
        // Grow the size and realloc the block to handle it.
        config->commandTableSize += 10;
        config->commands = (CommandStruct *) realloc (config->commands , config->commandTableSize * sizeof (CommandStruct));
    }
    
    // Store all the info for this command.
    int index = config->commandCount++;
    CommandStruct *table = config->commands;

    // Copy the command
    table[index].cmd = config->currentCommand;

    // Copy over each arg. We initially specify that we have no regex here, just in case we had
    // to realloc the config table above, which might flag us as having a regex when we don't.
    // This also keeps valgrind happy, since otherwise it complains about it not being initialized.
    // The only thing that doesn't get initialized here is the compiled regex structure, but this is
    // only ever accessed if haveRegex is TRUE, and in that case, it gets initialized during command
    // verification, so it's safe.
    table[index].arg              = (config->currentCommandArg == NULL ? NULL : strdup (config->currentCommandArg));
    table[index].keepIncomingCase = config->keepIncomingCase;
    table[index].spaceAction      = config->spaceAction;
    table[index].clobberIncoming  = config->clobberIncoming;
    table[index].haveRegex        = SDL_FALSE;

    // For the --to path, all commands except extract commands specify an NFI directory in --to.
    // So we fix up the --to path on all commands except extractions, and for extractions we just copy
    // what we are given.
    switch (table[index].cmd)
    {
        // These commands require the --to path to be whatever the user passed in.
        case CMD_EXTRACTFILES:
        case CMD_EXTRACTDIR:
        case CMD_EXTRACTDIRRECURSE:
        case CMD_EXTRACT:
        case CMD_EXTRACTRECURSE:
            table[index].toPath = (config->toPath == NULL ? NULL : strdup (config->toPath));
            break;

        // All other commands require a fixed --to path.
        default:
            table[index].toPath = fixCommandPath (config->toPath);
            break;
    }
    
    // For the --from path all commands except add commands specify an NFI directory
    // in --from. So we fix up the --from path on all commands except additions, and for
    // additions we just copy what we are given.
    switch (table[index].cmd)
    {
        // These commands require the --from path to be whatever the user passed in.
        case CMD_ADD:
        case CMD_ADDRECURSE:
        case CMD_ADDFILES:
        case CMD_ADDDIR:
        case CMD_ADDDIRRECURSE:
            table[index].fromPath = (config->fromPath == NULL ? NULL : strdup (config->fromPath));
            break;

        // All other commands require a fixed --from path
        default:
            table[index].fromPath = fixCommandPath (config->fromPath);
            break;
    }
}

/********************************************************************/

// Takes the command line parameters given to us and returns back a 
// configuration structure, or NULL on error, in which case an error will
// be printed.
//
// The config information is not fully parsed if a request for version
// information or help is given. If this returns success, you should check
// for one of those situations before continuing on.
ConfigInfo *openConfigInfo (int argc, char **argv) 
{
    // Allocate and initialize a config structure.
    ConfigInfo *config = (ConfigInfo *) malloc (sizeof (ConfigInfo));
    memset (config , 0 , sizeof (ConfigInfo));

    // Give the structure some commands to start with.
    config->commandCount     = 0;
    config->commandTableSize = 10;
    config->commands         = (CommandStruct *) malloc (config->commandTableSize * sizeof (CommandStruct));

    // By setting this to 0, the getopt_long function that we call will shut
    // the hell up when it encounters an error, allowing us to generate our
    // own error messages.
    opterr = 0;

    // This boolean flag is set to false at the beginning of each loop iteration
    // below, and set to true when --then is found. We use this to know if we
    // need an implict --then after the last command.
    SDL_bool lastOptWasThen = SDL_FALSE;

    // Keep looping forever. Eventually a call to getopt_long will return -1
    // to indicate that all of it's options are used up, and then we will break
    // manually.
    while (1) 
    {
        // Call getopt to parse out a value. It takes the argc/argv that we
        // get from main to start.
        //
        // Next is a list of short (1 character) options, in a string. If an
        // option is followed by a ':', it requires a value. Setting the very
        // first character of this string to a ':' character tells getopt that
        // it should distinguish between "this argument needs a value" and
        // "this argument is unknown". 
        //
        // Next we pass in an array of long options (see above) that defines
        // what arguments we understand.
        //
        // The last value is a pointer to an integer. If getopt_long recognizes
        // a long option successfully, this will get filled out to the value of
        // that options index in the passed in array. Note that if getopt recognizes
        // the option, but it requires a paramter and that parameter is missing,
        // this value will be left untouched, as if an error occured.
        //
        // getopt_long returns:
        //   -1 : There are no more arguments to process
        //    0 : A long argument was detected, it's flag pointer is non-NULL, so
        //        the value in the flag value field was stored into the flag pointer.
        //   ':': An option was found that needed an argument but it wasn't specified.
        //   '?': An unknown option was found.
        //   'c': (any character but '?' and ':') A short option was recognized, or a long option
        //        whose flag value is set to a character constant and had a NULL
        //        flag pointer.
        //    x : (any integer but 0) a long option was recognized and it's flag
        //        pointer was NULL, so the flag value was returned.
        //
        // On success (found an option and it had an argument if it needed it),
        // option_index (last parameter to getopt_long) is set to the index in 
        // the long options array, if it was a long option, or left alone otherwise.
        // This is ignored if it is NULL.
        //
        // If the recognized option gave an option, it is stored in the string
        // pointer optarg. optarg is NULL otherwise.
        //
        // On failure because of an unknown option or a missing option value,
        // the option character is stored in optopt for a single character
        // short option, or else optind is set to the argv pointer that was
        // being processed for a long argument. This allows you to display
        // what the missing argument is.
        //
        // getopt mutilates the argv[] pointers as it goes so that when it
        // is complete, all of the argv[] strings that did not contain option
        // text or option values are pushed to the end. It will set optind 
        // to be the index of the first string in argv[] of this nature, so
        // that you can collect the rest of the command line arguments.
        int option = getopt_long (argc, argv, ":f:culvVHhnbds:", long_options, NULL);

        // If getopt returns -1, we ran out of arguments. If that is the case, then
        // break out of the loop. Otherwise, we have a good argument, so reset the
        // flag that says the last option seen was --then.
        if (option == -1)
            break;
        else
            lastOptWasThen = SDL_FALSE;

        // Handle the return value.
        switch (option) 
        {
            // These options are someone telling us what file to work with.
            case CMD_FILE:
            case 'f':
                if (config->nfiFilename != NULL)
                {
                    fprintf (stderr , "ERROR: You may only specify one filename to work with!\n\n");
                    closeConfigInfo (config);
                    return NULL;
                }
                config->nfiFilename = strdup (optarg);
                break;

            // We will be creating a new file
            case CMD_CREATE:
            case 'c':
                config->isCreating = SDL_TRUE;
                break;

            // We will be updating an existing file
            case CMD_UPDATE:
            case 'u':
                config->isUpdating = SDL_TRUE;
                break;

            // We will be listing the contents of a file
            case CMD_LIST:
            case 'l':
                config->isListing = SDL_TRUE;
                break;

            // Verbosity flag
            case CMD_VERBOSE:
            case 'v':
                config->isVerbose = SDL_TRUE;
                break;

            // Debug flag
            case CMD_DEBUG:
            case 'd':
                config->isDebugging = SDL_TRUE;
                break;

            // Just test what we would do but don't do it
            case CMD_TESTING:
            case 'n':
                config->isTesting = SDL_TRUE;
                break;

            // Set the location where temporary files are created
            case CMD_TMPDIR:
                if (config->tmpDir != NULL)
                    fprintf (stderr , "Warning: --tmpdir already set, ignoring '%s'\n" , optarg);
                else
                    config->tmpDir = strdup (optarg);
                break;

            // Someone is asking us for our version. Set the flag and return success
            // right now.
            case CMD_VERSION:
            case 'V':
                config->showVersion = SDL_TRUE;
                return config;

            // Someone is asking us for help. Set the flag and return success right now.
            case CMD_HELP:
            case 'H':
                config->showHelp = SDL_TRUE;
                return config;

            // Someone telling us to keep the incoming case of things
            case CMD_KEEPCASE:
                config->keepIncomingCase = SDL_TRUE;
                break;

            // Someone telling us to convert to lowercase
            case CMD_LOWCASE:
                config->keepIncomingCase = SDL_FALSE;
                break;

            // Keep spaces in incoming files
            case CMD_KEEPSPACE:
                config->spaceAction = SPACES_KEEP;
                break;

            // Convert spaces in incoming files
            case CMD_CONVSPACE:
                config->spaceAction = SPACES_CONVERT;
                break;

            // Remove spaces in incoming files
            case CMD_DELSPACE:
                config->spaceAction = SPACES_REMOVE;
                break;

            // Allow newly added files to clobber existing files
            case CMD_CLOBBER:
                config->clobberIncoming = SDL_TRUE;
                break;

            // Files will not be added if a file with the same name already exists
            case CMD_NOCLOBBER:
                config->clobberIncoming = SDL_FALSE;
                break;

            // Someone wants a sort.
            case CMD_SORT:
            case 's':
                config->sortType = getSortStyle (optarg);
                break;

            // Use a descending sort instead of an ascending sort.
            case CMD_DESCEND:
                config->isDescendingSort = SDL_TRUE;
                break;

            // Turn on brief listing mode
            case CMD_BRIEF:
                config->useBriefListing = SDL_TRUE;
                break;

            // Disable size output for brief listings.
            case CMD_NOSIZE:
                config->omitSize = SDL_TRUE;
                break;

            // Turn on human-readable size information
            case CMD_HUMAN:
            case 'h':
                config->useHumanSizes = SDL_TRUE;
                break;

            // Specify the --from path for archive commands
            case CMD_FROM:
                if (config->fromPath != NULL)
                    free (config->fromPath);
                config->fromPath = strdup (optarg);
                break;

            // Specify the --to path for archive commands
            case CMD_TO:
                if (config->toPath != NULL)
                    free (config->toPath);
                config->toPath = strdup (optarg);
                break;

            // Called for --then, assumes that a command is complete and
            // adds it to the list.
            case CMD_THEN:
                // If there is no current command, then doing this will
                // have no effect.
                if (config->currentCommand != 0)
                    addCommandToConfig (config);
                else
                    fprintf (stderr , "Warning: Misplaced --then. Specify a command first.\n");

                // The last parsed option is now --then
                lastOptWasThen = SDL_TRUE;
                break;

            // These are all commands that we can actually execute
            // at runtime. We will store them and their parameters
            // and wait for a --then so that we know it's time to
            // do something.
            case CMD_ADD:
            case CMD_ADDRECURSE:
            case CMD_ADDFILES:
            case CMD_ADDDIR:
            case CMD_ADDDIRRECURSE:
            case CMD_REMOVE:
            case CMD_REMOVERECURSE:
            case CMD_REMOVEFILES:
            case CMD_REMOVEDIR:
            case CMD_REMOVEDIRRECURSE:
            case CMD_MOVE:
            case CMD_MOVERECURSE:
            case CMD_MOVEFILES:
            case CMD_MOVEDIR:
            case CMD_MOVEDIRRECURSE:
            case CMD_EXTRACTFILES:
            case CMD_EXTRACTDIR:
            case CMD_EXTRACTDIRRECURSE:
            case CMD_EXTRACT:
            case CMD_EXTRACTRECURSE:
                // Store this command value
                config->currentCommand = (CommandType) option;

                // If there was a command argument for the previous command,
                // get rid of it now
                if (config->currentCommandArg != NULL)
                    free (config->currentCommandArg);

                // Store this argument if there is one.
                if (optarg == NULL)
                    config->currentCommandArg = NULL;
                else
                    config->currentCommandArg = strdup (optarg);
                break;

            // An option that required an argument was found, but no argument was specified.
            case ':':
                // optopt is set to what getopt_long would return if the parameter was provided,
                // so use that to look up what option is missing.
                fprintf (stderr , "%s needs a parameter\n\n" , optionNameFromValue (optopt));
                break;

            // An option that is unknown was discovered. 
            case '?':
                // optopt is the character if it's a short option. If that is null, the unknown
                // option is a long option. In that case before getopt returns it increments
                // optind to be the next argv that it will process.
                if (optopt == '\0')
                    fprintf (stderr , "Unknown option \"%s\"\n\n" , argv[optind - 1]);
                else
                    fprintf (stderr , "Unknown option '-%c'\n\n" , optopt);
                break;

            // Catch everything else. 
            default:
                fprintf (stderr , "?? getopt_long returned unhandled value %d ??\n", option);
                break;
        }
    }

    // Argument parsing is complete. If the last argument passed wasn't --then,
    // and we have hit at least one command, then we want to presume that there
    // was a trailing --then on the command line. This means that you don't need
    // to put --then on the end of the command line for a single command, but 
    // if you do, you won't get a duplicate command.
    if (lastOptWasThen == SDL_FALSE && config->currentCommand != 0)
        addCommandToConfig (config);

    // When getopt returns -1. it sets optind to be the index of the first
    // argv[] string that it didn't use. We can use this to pick up the rest of the
    // text from the command line.
    if (optind < argc) 
    {
        // How many arguments are there to process?
        config->cmdlineFileCount = argc - optind;

        // Allocate an array of pointers.
        config->cmdlineFiles = (char * *) malloc (config->cmdlineFileCount * sizeof (char *));

        // Now duplicate all of those arguments 
        for (int i = 0 ; i < config->cmdlineFileCount ; i++)
            config->cmdlineFiles[i] = strdup (argv[optind + i]);
    }

    // If the config validates out, then return it back
    if (validateConfig (config) == SDL_TRUE)
        return config;

    // There is a problem, so close up the config and return an error.
    closeConfigInfo (config);
    return NULL;
}

/********************************************************************/

