/********************************************************************/

#include "Tool.h"

/********************************************************************/

// This takes an extract command and configuration information. If the
// --no-action/-n option is in effect, this will output text to display
// what the command would do.
static void printExtractCommand (CommandStruct *command , ConfigInfo *config)
{
    // If we are not just testing, do nothing here.
    if (config->isTesting == SDL_FALSE)
        return;

    // Based on what extract command this is, act accordingly.
    switch (command->cmd)
    {
        case CMD_EXTRACTFILES:
            printf ("-> %s: Files from the command line in the NFI at '%s' will be extracted to '%s'\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        case CMD_EXTRACTDIR:
            printf ("-> %s: All files in '%s' in the archive will be extracted to '%s'\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        case CMD_EXTRACTDIRRECURSE:
            printf ("-> %s: All files and subdirectories of '%s' in the archive will be extracted to '%s'\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        case CMD_EXTRACT:
            printf ("-> %s: All files in '%s' in the archive whose names match '%s' will be extracted to '%s'\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->arg , command->toPath);
            break;

        case CMD_EXTRACTRECURSE:
            printf ("-> %s: All files and subdirectories of '%s' in the archive whose names match '%s' will be extracted to '%s'\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->arg , command->toPath);
            break;

        default:
            break;
    }

    // If this is an option that takes command line files, display them now.
    if (command->cmd == CMD_EXTRACTFILES)
    {
        printf ("    -> ");

        for (int i = 0 ; i < config->cmdlineFileCount ; i++)
            printf ("%s " , config->cmdlineFiles[i]);

        printf ("\n");
    }
}

/********************************************************************/

// --extract-files: Find all files that exist in the archive in the
// --from path that are given on the command line, and write them
// to the real filesystem at --to.
void cmdExtractFiles (NFIFile *nfi , CommandStruct *command , ConfigInfo *config)
{
    // print out what we're going to be doing
    printExtractCommand (command , config);

    // Loop over all files that were given on the command line. For each one, construct
    // a full path inside the NFI archive, and then try to extract the file from it.
    for (int i = 0 ; i < config->cmdlineFileCount ; i++)
    {
        char thisFile[NFI_NAMELEN + 1];

        // Build a full filename, which consists of the from path, and the path given
        // on the command line (which may also contain path information, which will
        // be considered relative to fromPath)
        sprintf (thisFile , "%s%s" , command->fromPath , config->cmdlineFiles[i]);

        // Try to extract the file now. 
        nfi->pathStruct->extractFile (thisFile , command->toPath , command->clobberIncoming);
    }
    
    // All done. If we are outputting what we're doing, seperate this command from any that might follow
    if (config->isTesting)
        printf ("\n");
}

/********************************************************************/

// --extract-dir: Extract all files in the archive at --from
// to the real filesystem at --to. Only the files exactly in the --from
// path will be extracted.
//
// --extract-dir-recurse: Extract all files and directories in the NFI archive
// at --from to the real filesystem at --to. Any directories relative to
// --from will also be relative to --to when the files are extracted (so 
// that the directory structure inside the NFI archive is maintained).
//
// --extract: Compare the filenames of all files that exist in the
// path --from, and extract any that match to the real filesystem at --to.
//
// --extract-recurse: Compare all files that exist in the nfi archive --from
// path and below to the given regex (filenames only). If any match, extract
// them to the real filesystem at --to. Directory structure of extracted
// files is maintained so that any paths relative to --from will be relative
// to the --to path as well.
void cmdExtract (NFIFile *nfi , CommandStruct *command , ConfigInfo *config)
{
    // print out what we're going to be doing
    printExtractCommand (command , config);

    // Pretend that we won't recurse with this command.
    SDL_bool recurse = SDL_FALSE;

    // Check to see if the command we're executing is one of the ones that
    // optionally recurse, and set the recurse flag to true if it is.
    if (command->cmd == CMD_EXTRACTDIRRECURSE || command->cmd == CMD_EXTRACTRECURSE)
        recurse = SDL_TRUE;

    // Pretend that we have no regex
    regex_t *regex = NULL;

    // If this is a command that uses a regex, then set that up
    if (command->cmd == CMD_EXTRACT || command->cmd == CMD_EXTRACTRECURSE)
        regex = &command->compiledRegex;

    // Now, call into the NFI path structure and get it to extract the files for us. 
    nfi->pathStruct->extractFiles (command->fromPath , command->toPath , regex , recurse , command->clobberIncoming);

    // All done. If we are outputting what we're doing, seperate this command from any that might follow
    if (config->isTesting)
        printf ("\n");
}

/********************************************************************/


