/********************************************************************/

#include "Tool.h"

/********************************************************************/

// Construct
ScannedFiles::ScannedFiles (const char *path , regex_t *regexExpr , SDL_bool isVerbose , SDL_bool recurse)
{
    // Initialize the lists.
    maxNameSize = nameSize = 0;
    maxDirSize  = dirSize  = 0;
    names = dirs = NULL;
    rootDir = NULL;

    // Grow both lists to their initial sizes.
    growNameList ();
    growDirList ();    

    // Save the regex given, if any
    regex = regexExpr;

    // Save the root path so that as we are adding we know what to clip off of the names
    // we find.
    rootDir = strdup (path);

    // Calculate the length of the root directory we're searching,
    // plus 1 byte for a trailing / character. We use this to not
    // store the root directory in things that we store, to save
    // time and space.
    rootDirLen = strlen (rootDir) + 1;

    // Are we verbose?
    bVerbose = isVerbose;

    // Scan now. We want an initial scan of the rootDir provided.
    findFilesInPath (SDL_TRUE , recurse);
}

/********************************************************************/

// Take a list of pre-scanned files, relative to the current directory. Each name
// will be checked to see if it exists or not. If it is, it will be added. Any
// of the files specified that are paths will be recursed into.
ScannedFiles::ScannedFiles (const char **files, int count, SDL_bool isVerbose)
{
    // Initialize the lists.
    maxNameSize = nameSize = 0;
    maxDirSize  = dirSize  = 0;
    names = dirs = NULL;
    rootDir = NULL;
    regex = NULL;

    // Grow both lists to their initial sizes.
    growNameList ();
    growDirList ();    

    // Save the root path so that as we are adding we know what to clip off of the names
    // we find.
    rootDir = strdup (".");

    // Calculate the length of the root directory we're searching,
    // plus 1 byte for a trailing / character. We use this to not
    // store the root directory in things that we store, to save
    // time and space.
    rootDirLen = strlen (rootDir) + 1;

    // Are we verbose?
    bVerbose = isVerbose;

    // Add all files and paths given to the constructor.
    for (int i = 0 ; i < count ; i++)
        addText (rootDir , files[i]);

    // Scan now, skipping the scan of the rootDir provided because we
    // already have added the files we want to start with.
    findFilesInPath (SDL_FALSE , SDL_TRUE);
}

/********************************************************************/

// Destruct
ScannedFiles::~ScannedFiles (void)
{

    // Get rid of all names we may we storing.
    if (names)
    {
        for (int i = 0 ; i < nameSize ; i++)
            free (names[i]);

        free (names);
    }
        
    // Get rid of any directories we may we storing.
    if (dirs)
    {
        for (int i = 0 ; i < dirSize ; i++)
        {
            // If this path still exists, release it
            if (dirs[i] != NULL)
                free (dirs[i]);
        }

        free (dirs);
    }

    // Release the root directory
    if (rootDir)
        free (rootDir);
}

/********************************************************************/

// Grows the size of the name list
void ScannedFiles::growNameList (void)
{
    // Bump the max up
    maxNameSize += NAME_GROW_SIZE;

    // Re-allocate the list.
    names = (char * *) realloc (names , maxNameSize * sizeof (char *));
}

/********************************************************************/

// Grows the size of the dir list
void ScannedFiles::growDirList (void)
{
    // Bump the max up
    maxDirSize += DIR_GROW_SIZE;

    // Re-allocate the list.
    dirs = (char * *) realloc (dirs , maxDirSize * sizeof (char *));
}

/********************************************************************/

// Path is the path we are currently searching, and text is the name
// of a file/dir found in that path. If the name represents a directory,
// it is added to the list of directories that have been seen. If
// it is a file, it is stored in the name list.
//
// Names stored in the name list may be subject to a regex comparison
// if one was specified. In this case, only filenames that match the
// regex will be added.
//
// Any added filename does not store the root path, that is, the
// path that was initially passed to the constructor. This saves
// space since that would be redundant.
void ScannedFiles::addText (const char *path , const char *text)
{
    struct stat info;
    static char fullname[1024];

    // Combine the path and text together so that we get it's full
    // name.
    sprintf (fullname , "%s/%s" , path , text);

    // Try to stat the file. If it doesn't work, leave.
    if (lstat (fullname , &info) == -1)
        return;

    // If the file is a directory, we want to add it to the dir list.
    if (S_ISDIR (info.st_mode))
    {
        // If the dir list is not big enough, grow it
        if (dirSize == maxDirSize)
            growDirList ();

        // Store in the dir list.
        dirs[dirSize++] = strdup (fullname);
    }
    else if (S_ISREG (info.st_mode))
    {
        // Do we have a regex?
        if (regex)
        {
            // If the filename doesn't match the name we got,
            // return without doing anything.
            if (regexec (regex , text , 0 , 0 , 0) != 0)
                return;
        }

        // If the name list is not big enough, grow it
        if (nameSize == maxNameSize)
            growNameList ();

        // Store in the name list.
        names[nameSize++] = strdup (fullname + rootDirLen);
    }
}

/********************************************************************/

// Scan for files at the provided path. We call addText to handle
// the data we get.
void ScannedFiles::scanPath (const char *path)
{
    DIR *handle;
    struct dirent *found;

    // Try to open the directory for scanning. If it fails,
    // stop. We always display scan errors wether we are verbose
    // or not, since it usually means permissions are invalid and
    // the user might not realize it otherwise.
    handle = opendir (path);
    if (handle == NULL)
    {
        fprintf (stderr , "Unable to scan '%s'\n" , path);
        return;
    } else {
        // We opened the directory OK, so if we are verbose, say
        // that we're scanning it
        if (bVerbose)
            printf ("Scanning '%s'\n" , path);
    }

    // Read all things
    while ((found = readdir (handle)) != NULL)
    {
        // If we found "." or "..", then skip it. We don't want to recurse
        // down, and we already read the contents of ".".
        if (strcmp (found->d_name , ".") == 0 || strcmp (found->d_name , "..") == 0)
            continue;

        addText (path , found->d_name);
    }

    // Close it up now
    closedir (handle);
}

/********************************************************************/

// Scan files for inclusion in the list. If initialScan is TRUE, then
// we scan the rootDir given in the constructor and add all files and
// paths specified there. If initialScan is FALSE, this step is skipped
// and it is assumed that the internal lists have been filled out with
// some other information to begin with.
//
// If recurse is TRUE, all directories found in the initial scan (or placed
// in the directory list by code prior to this call) will be scanned the
// same as the rootDir was above.
void ScannedFiles::findFilesInPath (SDL_bool initialScan , SDL_bool recurse)
{
    if (initialScan)
    {
        // To start off, scan the initial directory that we were given.
        // any regular files found will be added to the list (optionally,
        // if they match the regex given the constructor). Any directories
        // found inside the search path will be added to the directory list.
        scanPath (rootDir);
    }

    // Should we recurse into the directories that we found?
    if (recurse)
    {
        // Now, scan all directories that may have been found. The
        // subdir scan will add all of the files in the provided path,
        // potentially adding new directories to the dir list as well.
        // When this is done, we have searched all paths available.
        //
        // As we finish with a directory, we release it's name, since
        // at that point we no longer need. For deep directory trees
        // this is a fairly substantial space savings.
        for (int i = 0 ; i < dirSize ; i++)
        {
            // Scan the path at this index. This might make dirSize
            // increse.
            scanPath (dirs[i]);

            // Free the path and mark it as NULL, since we're done
            // with it now.
            free (dirs[i]);
            dirs[i] = NULL;
        }
    }
}

/********************************************************************/

