/********************************************************************/

#include "Tool.h"

/********************************************************************/

// This special value is used in all of the sort routines. They
// multiply their return value by this number. When this is positive,
// the result is an ascending sort. When it's negative, the result is
// a descending sort.
static int sortModifier = 1;

/********************************************************************/

static int nfiNameSort (const void *a , const void *b)
{
    NFIParsedDirEntry *left = (NFIParsedDirEntry *) a;
    NFIParsedDirEntry *right = (NFIParsedDirEntry *) b;

    return strcmp (left->name , right->name) * sortModifier;
}

/********************************************************************/

static int nfiPathSort (const void *a , const void *b)
{
    NFIParsedDirEntry *left = (NFIParsedDirEntry *) a;
    NFIParsedDirEntry *right = (NFIParsedDirEntry *) b;

    // Check to see if the paths need sorting. If they're different,
    // then base the sort on that value.
    int compare = strcmp (left->path , right->path);
    if (compare != 0)
        return compare * sortModifier;

    // The paths are the same, so base the sort decision on the name.
    return strcmp (left->name , right->name) * sortModifier;
}

/********************************************************************/

static int nfiSizeSort (const void *a , const void *b)
{
    NFIParsedDirEntry *left = (NFIParsedDirEntry *) a;
    NFIParsedDirEntry *right = (NFIParsedDirEntry *) b;

    // Check to see if the lengths need sorting. If they're different,
    // then base the sort on that value.
    int compare = left->length - right->length;
    if (compare != 0)
        return compare * sortModifier;

    // The lengths are the same, so base the sort decision on the name.
    return (left->length - right->length) * sortModifier;
}

/********************************************************************/

static int nfiOffsetSort (const void *a , const void *b)
{
    NFIParsedDirEntry *left = (NFIParsedDirEntry *) a;
    NFIParsedDirEntry *right = (NFIParsedDirEntry *) b;

    // Sort based purely on offset. There will never be duplicate
    // offsets, so we can't sub-sort on name here.
    return (left->offset - right->offset) * sortModifier;
}

/********************************************************************/

// Takes a number and converts it into a string, justifying it to the
// width provided. If human is true, the number is converted to a human
// readable number (K/M/G), otherwise it is returned as a number.
//
// The buffer used is a static internal variable.
static const char *formatSize (int width , int size , SDL_bool human)
{
    static char buffer[59];
    char suffix = 'b';

    // If we were not told to make the number human readable,
    // format it using regular number logic and return it.
    if (human == SDL_FALSE)
    {
        sprintf (buffer , "%*d" , width , size);
        return buffer;
    }

    // Big enough to be a K?
    if (size > 1024)
    {
        size /= 1024;
        suffix = 'K';
    }

    // Enough left to be megs?
    if (size > 1024)
    {
        size /= 1024;
        suffix = 'M';
    }

    // Enough left to be gigs?
    if (size > 1024)
    {
        size /= 1024;
        suffix = 'G';
    }

    // Format it into a temporary string
    char temp[50];
    sprintf (temp , "%d %c" , size , suffix);

    // Format it into the buffer and return it.
    sprintf (buffer , "%*s" , width , temp);
    return buffer;
}

/********************************************************************/

// Return the number of columns in the current terminal.
static int getScreenCols (void)
{
    int cols=0;

    // Query the environment to see if it will tell us how many
    // columns the screen is.
    char *colVar = getenv ("COLUMNS");
    if (colVar != NULL)
        cols = atoi (colVar);

    // If it didn't work, default to 80
    if (cols == 0)
        cols = 80;

    // Return it, subtracting 1 so that we never write to the last
    // column of the screen if we can help it (which would force a line
    // break).
    return cols - 1;
}

/********************************************************************/

// Given a loaded file, determine how big the columns for the 4 standard
// pieces of data should be in the table view.
static void calculateTableSizes (NFIFile *file , int &nameSize , int &sizeSize , int &offsetSize , int &pathSize)
{
    // Try to get the number of columns on the screen. 
    int cols = getScreenCols ();

    // Set up the default minimum values for all of the columns.
    nameSize   = 4;
    sizeSize   = 8;
    offsetSize = 8;
    pathSize   = 4;

    // Subtract from the number of columns the space that it will take
    // to seperate 4 columns (3 spaces between columns), plus an extra
    // space on the right after the last column, so that if that column
    // goes to the full width it doesn't wrap. This comes out to be 
    // 10 characters: (counting all _ and | characters)
    //     -> COL1 _|_ COL2 _|_ COL3 _|_ COL4 _
    cols -= 10;

    // Now subtract the sizes for the size and offset columns, which
    // are always the default size no matter what.
    cols -= (sizeSize + offsetSize);

    // If there are no columns left (or we used to many) then return
    // right now. The terminal is too small to display things as it
    // is, so we will just use the defaults since it will look like
    // ass anyway.
    if (cols <= 0)
        return;

    // Determine the longest filename and directory name in the file.
    int fileMax = file->longestFilename;
    int pathMax = file->longestPathComponent;

    // If fileMax + pathMax is larger than the number of columns remaining,
    // try to split them in a sensible way and then return.
    if (fileMax + pathMax > cols)
    {
        // If there is not enough columns to at least make filenames 
        // line up, then split the remaining space between the two and
        // leave now.
        if (fileMax > cols)
        {
            nameSize = pathSize = cols / 2;
            return;
        }
    }

    // Here there is either enough space to give each the proper amount,
    // or at least enough so that the filenames will line up nice. Give
    // the names the space they need and allocate the rest to the path
    // section.
    nameSize = fileMax;
    pathSize = cols - fileMax;
}

/********************************************************************/

// Display a table separator line
static void nfiTableSep (int nameSize , int sizeSize , int offsetSize , int pathSize)
{
    int i;

    printf ("|-");
    for (i = 0 ; i < nameSize ; i++) printf ("-");
    printf ("-|-");
    for (i = 0 ; i < sizeSize ; i++) printf ("-");
    printf ("-|-");
    for (i = 0 ; i < offsetSize ; i++) printf ("-");
    printf ("-|-");
    for (i = 0 ; i < pathSize ; i++) printf ("-");
    printf ("\n");
}


/********************************************************************/

// Display the table header based on passed in sizes.
static void nfiTableHeader (int nameSize , int sizeSize , int offsetSize , int pathSize)
{
    printf ("| %-*s | %-*s | %-*s | %-*s\n" , 
       nameSize , "Name" , sizeSize , "Size" , offsetSize , "Offset" , pathSize , "Path");

    nfiTableSep (nameSize , sizeSize , offsetSize , pathSize);
}

/********************************************************************/

// Display an entry
static void nfiTableEntry (NFIParsedDirEntry *entry , int nameSize , int sizeSize , int offsetSize , int pathSize , SDL_bool human)
{
    printf ("| %-*s | %s | %*d | %-*s\n" , 
       nameSize   , entry->name , 
       formatSize (sizeSize , entry->length , human) , 
       offsetSize , entry->offset , 
       pathSize   , entry->path);
}

/********************************************************************/

// This method prints the contents of the given NFI file out in a verbose way.
static void printNFITable (NFIFile *file , ConfigInfo *config)
{
    // Set up
    printf ("\n\n");

    // Print out base information
    printf ("File          : %s\n" , file->name);
    printf ("NFI Version   : %d\n" , file->fileHeader.version);
    printf ("Dir Offset    : 0x%02X\n" , file->fileHeader.dirOffs);
    printf ("Files Stored  : %d file%s\n\n" , file->dirHeader.dirSize , (file->dirHeader.dirSize == 1 ? "" : "s"));
    printf ("Display Order : ");

    // If the sort should be a descending sort, set the sort modifier to reverse
    // the sign of all sort comparisons so that the sort comes out the way we want.
    if (config->isDescendingSort)
        sortModifier = -1;

    // Display sort type, then sort as appropriate
    switch (config->sortType)
    {
        case SORT_NATURAL:
            printf ("Internal storage order\n");
            break;

        case SORT_NAME:
            printf ("Sorted by Filename");
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiNameSort);
            break;

        case SORT_SIZE:
            printf ("Sorted by Size, then Filename");
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiSizeSort);
            break;

        case SORT_OFFSET:
            printf ("Sorted by Offset");
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiOffsetSort);
            break;

        case SORT_PATH:
            printf ("Sorted by Path, then Filename");
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiPathSort);
            break;
    }

    if (config->sortType != SORT_NATURAL)
    {
        if (config->isDescendingSort == SDL_TRUE)
            printf (" (in descending order)\n");
        else
            printf (" (in ascending order)\n");
    }

    // Print a table view
    printf ("\n\n");

    int nameSize,sizeSize,offsetSize,pathSize;
    calculateTableSizes (file , nameSize , sizeSize , offsetSize , pathSize);
    nfiTableHeader (nameSize , sizeSize , offsetSize , pathSize);

    for (Uint32 i = 0 ; i < file->dirHeader.dirSize ; i++)
        nfiTableEntry (&file->directory[i] , nameSize , sizeSize , offsetSize , pathSize , config->useHumanSizes);

    nfiTableSep (nameSize , sizeSize , offsetSize , pathSize);

    // All done
    printf ("\n\n");
}

/********************************************************************/

// Display an entry
static void nfiBriefEntry (NFIParsedDirEntry *entry , int nameSize , int sizeSize , SDL_bool human)
{
    // Consolidate the name and path back into 1 string.
    static char buffer[NFI_NAMELEN + 1];
    sprintf (buffer , "%s/%s" , entry->path , entry->name);

    // Output it.
    printf ("%-*s    %s\n" ,
       nameSize   , buffer , 
       formatSize (sizeSize , entry->length , human));
}

/********************************************************************/

// Display an entry
static void nfiBriefEntryNoSize (NFIParsedDirEntry *entry , int nameSize , int sizeSize)
{
    // Dump it
    printf ("%s/%s\n" , entry->path , entry->name);
}

/********************************************************************/

// This method prints the contents of the given NFI file out in a brief way.
static void printNFIBrief (NFIFile *file , ConfigInfo *config)
{
    // Sort the data according to what the user wants.
    switch (config->sortType)
    {
        case SORT_NATURAL:
            break;

        case SORT_NAME:
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiNameSort);
            break;

        case SORT_SIZE:
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiSizeSort);
            break;

        case SORT_OFFSET:
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiOffsetSort);
            break;

        case SORT_PATH:
            qsort (file->directory , file->dirHeader.dirSize , sizeof (NFIParsedDirEntry) , nfiPathSort);
            break;
    }

    // Calculate how many spaces it will take for the file+path portion of this operation.
    // This includes a single byte for a seperating / character between the two
    int nameSize = file->longestFilename + file->longestPathComponent+1;
    int sizeSize = 8;

    // Now display everything
    for (Uint32 i = 0 ; i < file->dirHeader.dirSize ; i++)
    {
        if (config->omitSize)
            nfiBriefEntryNoSize (&file->directory[i] , nameSize , sizeSize);
        else
            nfiBriefEntry (&file->directory[i] , nameSize , sizeSize , config->useHumanSizes);
    }
}

/********************************************************************/

// Print the contents of the NFI file.
void printNFIContents (NFIFile *file , ConfigInfo *config)
{
    // Do nothing on empty handle.
    if (file == NULL)
        return;
    
    // If we are supposed to be brief, use the brief printer. Otherwise,
    // use the verbose printer.
    if (config->useBriefListing)
        printNFIBrief (file , config);
    else
        printNFITable (file , config);
}

/********************************************************************/

