/********************************************************************/

#include "Tool.h"

/********************************************************************/

// This defines the size of the buffer that we use when copying data
// to the archive and when moving the completed archive around.
#define COPY_BUFFER_LEN 65535

/********************************************************************/

// Reads the NFI file header from the structure passed in. The return
// value tells you if it succeeded or not. On failure, the file is 
// closed before return.
static int readNFIFileHeader (NFIFile &info)
{
    // Try to read in the various parts of the main header one at a
    // time. All of the reads are enclosed in a try block. If there is
    // any error reading, an exception will be thrown that is caught 
    // at the bottom of the code, making it return SDL_FALSE.
    try
    {
        FILE *handle = info.fileHandle;

        // Get the header
        if (!fread (info.fileHeader.id , sizeof (info.fileHeader.id) , 1 , handle))
            throw 1;

        // Get the version
        if (!fread (&info.fileHeader.version , sizeof (info.fileHeader.version) , 1 , handle))
            throw 1;

        // Get reserved info
        if (!fread (&info.fileHeader.reserved , sizeof (info.fileHeader.reserved) , 1 , handle))
            throw 1;

        // Get directory offset.
        if (!fread (&info.fileHeader.dirOffs , sizeof (info.fileHeader.dirOffs) , 1 , handle))
            throw 1;

        // If the header ID and the file version aren't right, 
        // toss an error.
        if (info.fileHeader.version != NFI_VERSION)
            throw 1;

        if (strcmp (info.fileHeader.id , NFI_HEADERID) != 0)
            throw 1;
    }

    // This catch block will catch any thrown exception. It closes the
    // file handle that was passed in and then returns false.
    catch (...)
    {
        fclose (info.fileHandle);
        return 0;
    }

    // Otherwise, if we are here, it worked
    return 1;
}

/********************************************************************/

// Writes the NFI file header from the structure passed in. The return
// value tells you if it succeeded or not. On failure, the file is 
// closed before return.
static int writeNFIFileHeader (FILE *handle , NFIFile *info)
{
    // Try to write out the various parts of the main header one at a
    // time. All of the writes are enclosed in a try block. If there is
    // any error writing, an exception will be thrown that is caught 
    // at the bottom of the code, making it return SDL_FALSE.
    try
    {
        // Put the header
        if (!fwrite (info->fileHeader.id , sizeof (info->fileHeader.id) , 1 , handle))
            throw 1;

        // Put the version
        if (!fwrite (&info->fileHeader.version , sizeof (info->fileHeader.version) , 1 , handle))
            throw 1;

        // Put reserved info
        if (!fwrite (&info->fileHeader.reserved , sizeof (info->fileHeader.reserved) , 1 , handle))
            throw 1;

        // Put directory offset.
        if (!fwrite (&info->fileHeader.dirOffs , sizeof (info->fileHeader.dirOffs) , 1 , handle))
            throw 1;
    }

    // This catch block will catch any thrown exception. It closes the
    // file handle that was passed in and then returns false.
    catch (...)
    {
        fclose (handle);
        return 0;
    }

    // Otherwise, if we are here, it worked
    return 1;
}

/********************************************************************/

// Given a file header, this will seek to the directory header and
// read it in. The header must have been read first since it contains
// the location of the directory. The file is closed before return if 
// it fails.
static int readNFIDirHeader (NFIFile &info)
{
    // Try to read in the parts of the directory header. Any errors 
    // causes an exception to be thrown which is caught below.
    try
    {
        FILE *handle = info.fileHandle;

        // Seek to where the directory is.
        if (fseek (handle , info.fileHeader.dirOffs , SEEK_SET))
            throw 1;

        // Read directory size (in entries)
        if (!fread (&info.dirHeader.dirSize , sizeof (info.dirHeader.dirSize) , 1 , handle))
            throw 1;

        // Get reserved info
        if (!fread (&info.dirHeader.reserved , sizeof (info.dirHeader.reserved) , 1 , handle))
            throw 1;
    }

    // This catch block will catch any thrown exception. It closes the 
    // file handle that was passed in and then returns false.
    catch (...)
    {
        fclose (info.fileHandle);
        return 0;
    }

    // Otherwise, if we are here, it worked
    return 1;
}

/********************************************************************/

// Assuming that the file header has just been written out, this will
// write out the directory header. 
//
// The directory size information may be incorrect if there are more
// entries in the directory than will be written to the disk.
static int writeNFIDirHeader (FILE *handle , NFIFile *info)
{
    // Try to write out the parts of the directory header. Any errors 
    // causes an exception to be thrown which is caught below.
    try
    {
        // Put directory size (in entries)
        if (!fwrite (&info->dirHeader.dirSize , sizeof (info->dirHeader.dirSize) , 1 , handle))
            throw 1;

        // Put reserved info
        if (!fwrite (&info->dirHeader.reserved , sizeof (info->dirHeader.reserved) , 1 , handle))
            throw 1;
    }

    // This catch block will catch any thrown exception. It closes the 
    // file handle that was passed in and then returns false.
    catch (...)
    {
        fclose (handle);
        return 0;
    }

    // Otherwise, if we are here, it worked
    return 1;
}

/********************************************************************/

// Takes a raw entry as read from the file and processes the data
// somewhat to split it out into the way we want it stored for each
// access, such as splitting the name apart into the path it was stored
// in and the name of the file, etc.
static void parseEntry (NFIDirEntry *raw , NFIParsedDirEntry *parsed)
{
    // Make sure the parsed entry is entirely initialized
    memset (parsed , 0 , sizeof (NFIParsedDirEntry));

    // Copy the easy stuff over.
    parsed->offset = raw->offset;
    parsed->length = raw->length;
    memcpy (parsed->reserved , raw->reserved , sizeof (parsed->reserved));

    // Find the last path separator in the raw name.
    char *lastPath = strrchr (raw->name , '/');

    // Copy the data over. This mugs the data in the raw name, but that's
    // ok because it is not needed after this.
    if (lastPath == NULL)
    {
        // No path seperator, so everything is the name and "/" is the
        // path.
        strcpy (parsed->path , "/");
        strcpy (parsed->name , raw->name);
    }
    else
    {
        // The character after the last path separator
        // starts the name. Copy it over.
        strcpy (parsed->name , lastPath + 1);

        // Increment lastPath by 1 so that it is past the seperating
        // path character, then clobber it and copy the resulting name
        // field ito the path.
        lastPath++;
        *lastPath = '\0';
        strcpy (parsed->path , raw->name);
    }
}

/********************************************************************/

// Assuming the directory header has just been read out of the file,
// this will load in the directory entries for the file. A dynamic
// array is loaded to store them all. As always, on error the file is 
// closed before return.
static int readNFIDirEntries (NFIFile &info)
{
    // Initialize stuff.
    info.longestFilename = 0;
    info.longestPathComponent = 0;

    // This here loop that reads the entries are in a try block. Any
    // file error will throw an exception that is later caught to 
    // clean things up.
    try
    {
        // We read into this entry
        NFIDirEntry thisEntry;

        // Create an array that will contain the directory contents.
        info.directory = (NFIParsedDirEntry *) malloc (info.dirHeader.dirSize * sizeof (NFIParsedDirEntry));

        // Loop reading all of the directory entries
        for (Uint32 index = 0 ; index < info.dirHeader.dirSize ; index++)
        {
            FILE *handle = info.fileHandle;
            Uint32 thisFile,thisPath;

            // Get the filename.
            if (!fread (thisEntry.name , sizeof (thisEntry.name) , 1 , handle))
                throw 1;

            // Read it's offset into the file
            if (!fread (&thisEntry.offset , sizeof (thisEntry.offset) , 1 , handle))
                throw 1;

            // Read it's length in the file
            if (!fread (&thisEntry.length , sizeof (thisEntry.length) , 1 , handle))
                throw 1;

            // Read reserved information
            if (!fread (&thisEntry.reserved , sizeof (thisEntry.reserved) , 1 , handle))
                throw 1;

            // Parse the entry out
            parseEntry (&thisEntry , &info.directory[index]);

            // Get the sizes and see if this is longer or what.
            thisFile = strlen (info.directory[index].name);
            thisPath = strlen (info.directory[index].path);

            if (thisFile > info.longestFilename)
                info.longestFilename = thisFile;

            if (thisPath > info.longestPathComponent)
                info.longestPathComponent = thisPath;
        }
    }

    // This catch block will catch any thrown exception. It closes the 
    // file handle that was passed in and then returns false.
    catch (...)
    {
        free (info.directory);
        info.directory = NULL;
        fclose (info.fileHandle);
        return 0;
    }

    // Otherwise, if we are here, it worked
    return 1;
}

/********************************************************************/

// This writes out the directory of the given NFI file. All entries
// that are flagged as removed are not written. The directory size
// is returned on success, or -1 on error. If we return error, we close
// the file before returning.
//
// This is used to write out stub entries (before sizes/offsets may be known)
// and also to write out the updated entries after the archive construction is
// complete.
static Sint32 writeNFIDirEntries (FILE *handle , NFIFile *info)
{
    // Initialize the count of directory entries written out.
    Uint32 filesInArchive = 0;

    // This here loop that reads the entries are in a try block. Any
    // file error will throw an exception that is later caught to 
    // clean things up.
    try
    {
        // Loop writing all of the directory entries that are not flagged
        // as removed.
        for (Uint32 index = 0 ; index < info->dirHeader.dirSize ; index++)
        {
            // If this directory entry is flagged as removed, then do not
            // write it.
            if (info->directory[index].type == NFI_REMOVED_FILE)
                continue;

            // Not removed, so we're going to write it out. Count it as an 
            // entry in the directory.
            filesInArchive++;

            // Put the filename. We have to join it together with the path, since
            // at load time we parsed it out.
            char fullName[NFI_NAMELEN];
            memset (fullName , 0 , sizeof (fullName));
            sprintf (fullName , "%s%s" , info->directory[index].path , info->directory[index].name);
            if (!fwrite (fullName , sizeof (fullName) , 1 , handle))
               throw 1;
              
            // Write out the offset (this is most likely bogus but it takes up the space
            // we need).
            if (!fwrite (&info->directory[index].offset , sizeof (info->directory[index].offset) , 1 , handle))
                throw 1;

            // Write it's length in the file (again, this is may be bogus for newly added files)
            if (!fwrite (&info->directory[index].length , sizeof (info->directory[index].length) , 1 , handle))
                throw 1;

            // Write out reserved information
            if (!fwrite (&info->directory[index].reserved , sizeof (info->directory[index].reserved) , 1 , handle))
                throw 1;
        }
    }

    // This catch block will catch any thrown exception. It closes the 
    // file handle that was passed in and then returns false.
    catch (...)
    {
        fclose (handle);
        return -1;
    }
    
    // If we are here, it worked. Return the new directory size.
    return filesInArchive;
}

/********************************************************************/


// Reads in all of the directory and file informtion for the given NFI
// file. Returns a pointer to a dynmaically allocated NFIFile 
// structure on success or null on error.
NFIFile *openNFIFile (ConfigInfo *config)
{
    NFIFile temp;

    // Null out the entries in the temporary structore.
    memset (&temp , 0 , sizeof (temp));

    // Set into it the name of the NFI file
    strcpy (temp.name , config->nfiFilename);

    // Try to open the file.
    temp.fileHandle = fopen (config->nfiFilename , "rb");

    // If it didn't work, return NULL
    if (temp.fileHandle == NULL)
        return NULL;

    // All of the functions below here that read bits of the headers 
    // will close the file if an error happens.

    // Try to get the header. Return NULL on failure.
    if (!readNFIFileHeader (temp))
        return NULL;

    // Now try to get the Directory header. Return NULL on failure.
    if (!readNFIDirHeader (temp))
        return NULL;

    // Try to read all of the directory entries. Return NULL 
    // on failure.
    if (!readNFIDirEntries (temp))
        return NULL;

    // If we get here, everything must have worked out OK. In that
    // case, create a new dynamic node and copy this info over to it, 
    // then return it.
    NFIFile *retVal = (NFIFile *) malloc (sizeof (NFIFile));
    memcpy (retVal , &temp , sizeof (NFIFile));

    // Create a path struct from the return value before we leave.
    retVal->pathStruct = new NFIPathStructure (retVal , config->isVerbose);

    return retVal;
}

/********************************************************************/

// Allocate and create a ne NFI file with the given name. 
//
// This will open the file for writing, truncating any existing file 
// that may exist.
//
// In this case, the file will be empty, so there will be no need to
// pull information from it in order to create the new file. However,
// we leave the file open until we are ready to create a new file. This
// makes sure that we can actually access the file when the time comes,
// so we don't waste time with other operations only to find we can't
// write the file we want.
//
// Returns NULL on error (unable to create the file named).
NFIFile *createNFIFile (ConfigInfo *config)
{
    // Open the file for writing
    FILE *handle = fopen (config->nfiFilename , "wb");
    if (handle == NULL)
        return NULL;

    // If we get here, allocate a dynamic structure, fill it out and then
    // return it.
    NFIFile *retVal = (NFIFile *) malloc (sizeof (NFIFile));
    memset (retVal , 0 , sizeof (NFIFile));

    // Store the name and file handle.
    strcpy (retVal->name , config->nfiFilename);
    retVal->fileHandle = handle;

    // Flag this as a new file, so that when it comes time to write out
    // the new file, we can just use this handle instead of creating
    // one of our own.
    retVal->newFile = SDL_TRUE;

    // Count creating the file as a change so that when the file is closed,
    // as long as there were no errors the file will be saved out empty.
    retVal->changeCount = 1;

    // Fill out the header
    strcpy (retVal->fileHeader.id , NFI_HEADERID);
    retVal->fileHeader.version = NFI_VERSION;
    retVal->fileHeader.dirOffs = sizeof (retVal->fileHeader.id) + 
                                 sizeof (retVal->fileHeader.version) + 
                                 sizeof (retVal->fileHeader.reserved) + 
                                 sizeof (retVal->fileHeader.dirOffs);

    // Create a path struct from the return value before we leave.
    retVal->pathStruct = new NFIPathStructure (retVal , config->isVerbose);

    // Directory header, directory, and the other things not specified above
    // have been initialize to proper values for us by the memset above.
    // Return now.
    return retVal;
}

/********************************************************************/

// Assumes that the file represented by the passed in NFIFIle structure
// has had the path portion of it's loaded directory modified but nothing
// else. We close and re-open the file for writing, then write out a new
// directory that takes the new paths into account.
static int rewriteNFIDirectory (NFIFile *file , ConfigInfo *config)
{
    // Close out the handle in the file, if it hasn't been done already.
    if (file->fileHandle != NULL)
        fclose (file->fileHandle);

    // Now, reopen the archive for writing. We use r+ here so that we get
    // reading and writing, so that we won't trunacate the existing file.
    FILE *nfi = fopen (file->name , "r+b");
    if (nfi == NULL)
    {
        fprintf (stderr , "Unable to open '%s' for writing!\n\n" , file->name);
        return EXIT_FAILURE;
    }

    // If we're verbose, tell the user what we're doing,
    if (config->isVerbose)
        printf ("Rewriting NFI directory header in '%s'\n" , file->name);

    // To start with, we need to seek to the location where the directory
    // starts. This is the location of the directory header as specified in
    // the file header, plus the size of the elements of the directory header,
    // which will put us at the first byte of the first actual directory entry.
    fseek (nfi , file->fileHeader.dirOffs , SEEK_SET);
    fseek (nfi , sizeof (file->dirHeader.dirSize) + sizeof (file->dirHeader.reserved) , SEEK_CUR);

    // Now we can loop over all directory entries and write them out. For each one, we will
    // write out a new name portion, then skip over the other fields.
    for (Uint32 i = 0 ; i < file->dirHeader.dirSize ; i++)
    {
        char newName[NFI_NAMELEN];

        // Put the name and path of the parsed entry in the directory back into one field,
        // then write it out.
        memset (newName , 0 , sizeof (newName));
        sprintf (newName , "%s%s" , file->directory[i].path , file->directory[i].name);

        // Write it out
        fwrite (newName , sizeof (newName) , 1 , nfi);

       // Now skip over the other fields.
       fseek (nfi , sizeof (file->directory[i].offset) , SEEK_CUR);
       fseek (nfi , sizeof (file->directory[i].length) , SEEK_CUR);
       fseek (nfi , sizeof (file->directory[i].reserved) , SEEK_CUR);
    }

    // Done now, close up the handle.
    fclose (nfi);
    if (config->isVerbose)
        printf ("Done!\n\n");

    return EXIT_SUCCESS;
}

/********************************************************************/

// Copy the actual contents of a file into the given file handle.
// If the entry passed in is flagged for removal, we return success
// right away. Otherwise, we copy the data either from the NFI archive
// passed in, or from some file on the filesystem. The entry is updated
// with the length of the new file and it's position in the new file.
static int writeNFIArchiveFile (FILE *handle , NFIFile *archive , NFIParsedDirEntry *entry , Uint8 *copyBuffer , int copyBufferSize)
{
    // If this file is supposed to be removed, then return success now.
    if (entry->type == NFI_REMOVED_FILE)
        return 1;

    // Determine the handle of the file we should read from. If the type of this entry is
    // added, then we need to actually try and open the source file.  Otherwise, the data
    // is going to come from our archive.
    FILE *dataHandle;
    if (entry->type == NFI_ADDED_FILE)
    {
        dataHandle = fopen (entry->sourceName , "rb");
        if (dataHandle == NULL)
        {
            fprintf (stderr , "Unable to open '%s' for addition to the archive!\n\n" , entry->sourceName);
            fclose (handle);
            return 0;
        }
    }
    else
    {
        dataHandle = archive->fileHandle;
    }

    // Set up the position in the new archive where this file will be stored
    entry->offset = ftell (handle);

    // Set up the size of the file. If the file type is added, we need to scan the file
    // we just opened to see it's length. Otherwise, it's already correct.
    if (entry->type == NFI_ADDED_FILE)
    {
        fseek (dataHandle , 0 , SEEK_END);
        entry->length = ftell (dataHandle);
        fseek (dataHandle , 0 , SEEK_SET);
    }

    // Copy all of the bytes over from one handle to the other.
    SDL_bool readError,writeError;
    copyFileData (dataHandle , handle , entry->length , copyBuffer , copyBufferSize , &readError , &writeError);

    // Here the file is either copied, or there was an error. To start with,
    // if we opened a handle for reading, close it. We leave it alone if the
    // file we were reading from is the archive.
    if (entry->type == NFI_ADDED_FILE)
        fclose (dataHandle);

    // Was there an error?
    if (readError || writeError)
    {
        // Display an error and return failure
        fprintf (stderr , "%s error saving file %s/%s\n" , (readError == SDL_TRUE ? "Read" : "Write") , entry->path , entry->name);
        return 0;
    }

    // Worked
    return 1;
}

/********************************************************************/

// Generate a temorary name for the file we're writing.
static void getTmpNFIName (char *buffer , ConfigInfo *config)
{
    sprintf (buffer , "%s/nfitool_tmpfile_%d.nfi" , config->tmpDir , getpid ());
}

/********************************************************************/

// Move an archive from one filename to another, like the rename 
// command. We try to complete the operation as fast as possible. 
static int moveArchive (const char *tmpName , const char *newName)
{
    // First of all, try to work the rename operation. If both
    // files are part of the same filesystem, this completes more
    // or less instantly.
    if (rename (tmpName , newName) == 0)
        return EXIT_SUCCESS;

    // The rename command failed. We know that we just successfully
    // created an archive from an existing archive, so if this fails
    // it must mean that the files are on different mounted filesystems.
    // In that case, we have to copy ourselves. Start by trying to open
    // the files we need.
    FILE *inFile = fopen (tmpName , "rb");
    if (inFile == NULL)
    {
        fprintf (stderr , "Unable to open file '%s' for reading during archive copy operation!\n\n" , tmpName);
        return EXIT_FAILURE;
    }

    // This operation will clobber the existing archive.
    FILE *outFile = fopen (newName , "wb");
    if (outFile == NULL)
    {
        fclose (inFile);
        fprintf (stderr , "Unable to open file '%s' for writing during archive copy operation!\n\n" , newName);
        return EXIT_FAILURE;
    }

    // Determine the filelength of the in file.
    fseek (inFile , 0 , SEEK_END);
    Uint32 fileLength = ftell (inFile);
    fseek (inFile , 0 , SEEK_SET);

    // Allocate a buffer to copy the data over
    Uint8 *buffer = (Uint8 *) malloc (COPY_BUFFER_LEN);

    // Copy the file over.
    SDL_bool readError,writeError;
    copyFileData (inFile , outFile , fileLength , buffer , COPY_BUFFER_LEN , &readError , &writeError);

    // Done with the buffer and the files
    fclose (inFile);
    fclose (outFile);
    free (buffer);
    
    // If there was a read or write error, then tell the user
    if (readError | writeError)
    {
        fprintf (stderr , "%s error copying archive from '%s' to '%s'\n" , (readError == SDL_TRUE ? "Read" : "Write") , tmpName , newName);
        fprintf (stderr , "File '%s' left behind for backup purposes.\n\n" , tmpName);
        return EXIT_FAILURE;
    }

    // It worked, so remove the temporary file before we return
    unlink (tmpName);

    return EXIT_SUCCESS;    
}

/********************************************************************/

// Given an NFIFile with some changes, save it back to disk. This works
// wether the file was being updated or newly created. This always
// writes out a bare file, even if it's empty.
static int saveNFIFile (NFIFile *file , ConfigInfo *config)
{
    // We are saving the file. If this isn't a new file and all of the
    // changes to the file were move commands, we can speed the operation
    // up by just re-writing the directory structure of the file instead
    // of doing a full write.
    if (file->newFile == SDL_FALSE && file->movesOnly)
        return rewriteNFIDirectory (file , config);

    // Start up
    if (config->isVerbose)
        printf ("Saving NFI file '%s'\n" , file->name);

    // To begin with, we must try to open a file for writing. To do
    // that, we first need a name for the file in question.
    char tmpName[NFI_NAMELEN];
    getTmpNFIName (tmpName , config);

    // Now try to open
    FILE *newNFI = fopen (tmpName , "wb");
    if (newNFI == NULL)
    {
        fprintf (stderr , "Unable to create temporary NFI file at '%s'\n\n" , tmpName);
        return EXIT_FAILURE;
    }

    // To start with, write out the file header. If this fails, it will close
    // the file before it returns failure, so we can just display an error and
    // return failure after unlinking the partially created file.
    if (writeNFIFileHeader (newNFI , file) == 0)
    {
        fprintf (stderr , "Error writing NFI header!\n\n");
        unlink (tmpName);
        return EXIT_FAILURE;
    }

    // Now, write out the directory header. Note that the directory size might
    // be incorrect here, if we removed files or added files but noclobber made
    // us add not as many files as we thought, etc. We'll come back later and
    // fix this.
    if (writeNFIDirHeader (newNFI , file) == 0)
    {
        fprintf (stderr , "Error writing NFI directory header!\n\n");
        unlink (tmpName);
        return EXIT_FAILURE;
    }

    // Write out the actual directory entries. On success, this will return the number
    // of entries actually written out, so that we can determine how many entries
    // will be added to the new file. On error, -1 is returned and the file is closed for us.
    Sint32 newDirSize = writeNFIDirEntries (newNFI , file);
    if (newDirSize == -1)
    {
        fprintf (stderr , "Error writing NFI directory entries!\n\n");
        unlink (tmpName);
        return EXIT_FAILURE;
    }

    // Now, we can write out each file in the archive. This may involve copying 
    // data from some external file into a new file, or it might involve copying
    // the data from the current NFI archive to the new one. ALong the way, the
    // directory entry needs to be updated with the size of the file being copied (if needed)
    // and it's new offset in the new file.
    Uint8 *copyBuffer = (Uint8 *) malloc (COPY_BUFFER_LEN);
    for (Uint32 index = 0 ; index < file->dirHeader.dirSize ; index++)
    {
        // Try to write out this file. If it doesn't work, return failure.
        if (writeNFIArchiveFile (newNFI , file , &file->directory[index] , copyBuffer , COPY_BUFFER_LEN) == 0)
        {
            fprintf (stderr , "Error writing NFI file contents!\n\n");
            unlink (tmpName);
            free (copyBuffer);
            return EXIT_FAILURE;
        }
    }

    // Done with the copy buffer now.
    free (copyBuffer);

    // Go back and write out the directory size byte in the directory header to represent how
    // many files we actually saved into the archive. Seek from the start of the file to just
    // past the file header, which is the proper location for the directory size.
    fseek (newNFI , sizeof (file->fileHeader.id) + 
                    sizeof (file->fileHeader.version) +
                    sizeof (file->fileHeader.reserved) +
                    sizeof (file->fileHeader.dirOffs) , SEEK_SET);
    if (fwrite (&newDirSize , sizeof (file->dirHeader.dirSize) , 1 , newNFI) != 1)
    {
        fprintf (stderr , "Error updating directory size!\n\n");
        fclose (newNFI);
        unlink (tmpName);
        return EXIT_FAILURE;
    }

    // Go back and write out the updated size and offset of all files added to the archive. The above
    // write operation put the file handle at the first byte of the reserved section of the directory
    // header, so seek past that. Once that is done, loop over all directory entries and rewrite them
    // to update the location information in them.
    fseek (newNFI , sizeof (file->dirHeader.reserved) , SEEK_CUR);
    if (writeNFIDirEntries (newNFI , file) == -1)
    {
        fprintf (stderr , "Error updating directory size!\n\n");
        unlink (tmpName);
        return EXIT_FAILURE;
    }

    // We're done. Close up the temporary file, then close out the handle in the original
    // archive, then move the temporary file to the original file.
    fclose (newNFI);
    fclose (file->fileHandle);
    file->fileHandle = NULL;

    // Try to move the temporary file to the existing file to replace the archive with it's
    // changed component. Our return value is whatever this returns.
    return moveArchive (tmpName , config->nfiFilename);
}

/********************************************************************/

// Print out the directory of the given NFI file, which shows any
// operations that may have been done on it, and how they will be saved.
void printNFIDirectory (NFIFile *file)
{
    // Display some state info
    printf ("\n-------------------------------------\n\n");
    printf ("Is new file : %s\n" , file->newFile == SDL_TRUE ? "YES" : "NO");
    printf ("Change count: %d\n" , file->changeCount);
    printf ("Moves Only  : %s\n" , file->movesOnly == SDL_TRUE ? "YES" : "NO");
    printf ("\n-------------------------------------\n\n");

    // Display the path contents
    file->pathStruct->displayContents ();

    // Display the directory we're about to work with
    printf ("\n------------- NFI Directory Structure -------------\n\n");

    for (Uint32 i = 0 ; i < file->dirHeader.dirSize ; i++)
    {
        printf ("%2d : " , i);

        switch (file->directory[i].type)
        {
            case NFI_UNTOUCHED:     printf ("    "); break;
            case NFI_ADDED_FILE:    printf ("(+) "); break;
            case NFI_REMOVED_FILE:  printf ("(-) "); break;
            case NFI_MOVED_FILE:    printf ("(~) "); break;
        }

        printf ("%-20s %-20s %-20s size=%8d offs=%8d (%p)\n" , 
           file->directory[i].name , 
           file->directory[i].path , 
           file->directory[i].sourceName ,
           file->directory[i].length ,
           file->directory[i].offset ,
           &(file->directory[i]));
    }

    printf ("\n---------------------------------------------------\n\n");
}

/********************************************************************/

// Close and get rid of a loaded NFI file. File can be NULL, in which
// case nothing happens. If the file has had any changes made to it
// since it was opened/created, it will be re-written. 
//
// Changes are only written out if noSave is false. If noSave is true,
// then no save operation occurs. This is used by the --no-action/-n
// command to stop any changes from being commited.
//
// If isDebugging is set to YES, the directory of the file being output
// is displayed, along with the internal path structure of the archive.
// This can help debug problems with commands interacting improperly.
//
// The return value should be EXIT_SUCCESS or EXIT_FAILURE to indicate
// if the close worked or not (particularly in the case of a save operation)
int closeNFIFile (NFIFile *file , SDL_bool noSave , ConfigInfo *config)
{
    int retVal=EXIT_SUCCESS;

    // Print out the directory as it stands right now, if we are debugging
    if (config->isDebugging)
        printNFIDirectory (file);

    // Do nothing if the handle is invalid.
    if (file == NULL)
        return EXIT_SUCCESS;

    // If there have been any changes made to the file, save them.
    if (file->changeCount != 0 && noSave == SDL_FALSE)
        retVal = saveNFIFile (file , config);

    // Close the handle if it's open
    if (file->fileHandle != NULL)
        fclose (file->fileHandle);

    // Get rid of any loaded directory.
    if (file->directory != NULL)
        free (file->directory);

    // If there is a path struct attached to us, blow it away now
    if (file->pathStruct)
        delete file->pathStruct;

    // Release the memory used to store the file data.
    free (file);

    // Return success unconditionally if we didn't save anything,
    // or the status of the save operation otherwise.
    return retVal;
}

/********************************************************************/

