/********************************************************************/

#include "Tool.h"

/*************************************************************************/

// Print the version number of the program
static void printVersion (void)
{
    printf ("nfitool 1.0.0\n");
    printf ("Copyright (C) 2004 Terence Martin\n");
    printf ("\n");
}

/********************************************************************/

// List the contents of an archive.
static int listContents (ConfigInfo *config)
{
    // Presume success
    int retVal = EXIT_SUCCESS;

    // Try to open the file
    NFIFile *file = openNFIFile (config);

    // If it didn't work, complain
    if (file == NULL)
    {
        fprintf (stderr , "Unable to open '%s'\n\n" , config->nfiFilename);
        retVal = EXIT_FAILURE;
    }

    // Try to print the contents of the file, then close it. These will behave
    // properly even if they are given a NULL file.
    printNFIContents (file , config);
    
    closeNFIFile (file , SDL_TRUE , config);

    return retVal;
}

/********************************************************************/

// Create/Update an archive.
static int createOrUpdate (ConfigInfo *config)
{
    // Special case. If we are just testing what would be done, but
    // there were no commands specified, then say that. Otherwise
    // things will be silent (mostly), and won't produce the trace
    // output that we expect.
    if (config->isTesting && config->commandCount == 0)
    {
        printf ("-> No commands given, ");

        if (config->isUpdating)
            printf ("nothing would be done\n\n");
        else
            printf ("an empty NFI archive would be created\n\n");

        return EXIT_SUCCESS;
    }

    // If there are no commands and we are updating, then complain
    // and return. This allows us to create an empty NFI file if we
    // want by just stating that we're going to create it but not
    // specifying any add operations and so on.
    if (config->commandCount == 0 && config->isUpdating)
    {
        // Consider this a failure because someone called us wrong.
        fprintf (stderr , "Nothing to do!\n\n");
        return EXIT_FAILURE;
    }

    // Try to open the NFI file that we were given.
    NFIFile *file = openNFIFile (config);

    // If we were trying to create a file and this open worked,
    // then we are unhappy because the file exists. If we were trying
    // to update a file and the open failed, then we are unhappy because
    // the file did not exist. If either of these two situations have
    // occured, close the file (as appropriate), complain and return.
    if (config->isCreating)
    {
        // Trying to create. If the open worked, complain and leave.
        if (file != NULL)
        {
            fprintf (stderr , "Cowardly refusing to clobber an existing archive!\n\n");
            closeNFIFile (file , SDL_TRUE , config);
            return EXIT_FAILURE;
        }

        // Create the file we want to create
        file = createNFIFile (config);

        // Complain if we couldn't create it.
        if (file == NULL)
        {
            fprintf (stderr , "Unable to create new archive '%s'!\n\n" , config->nfiFilename);
            return EXIT_FAILURE;
        }
    }
    else
    {
        // Trying to update. If the open failed, complain and leave.
        if (file == NULL)
        {
            fprintf (stderr , "Unable to open '%s' for update!\n\n" , config->nfiFilename);
            return EXIT_FAILURE;
        }
    }

    // Using the file we have, perform all needed commands.
    performCommands (file , config);

    // Check the config. If we were running in a "What would you do?" mode and
    // this is a new file we just created, then we will force the change count to 
    // be 0, pre-close the file, and then unlink it. This is because if you specify
    // create, the file is created right away. the closeNFIFile function is smart enough
    // to not try to close the file if we set the handle to NULL, and it won't try to
    // write changes if the change count is 0.
    if (file->newFile && config->isTesting)
    {
        // Close the file, then remove it. We have to set the handle in the file
        // to NULL so that the closeNFIFile function doesn't try to close it.
        if (file->fileHandle != NULL)
        {
            fclose (file->fileHandle);
            file->fileHandle = NULL;
        }

        // Remove the file now.
        unlink (file->name);

        // Force the change count to 0, just in case.
        file->changeCount = 0;
    }

    // Close the NFI file now. If there were any operations done to move/add/delete
    // things from the file, and we are not testing, this will save them before 
    // it closes things out.
    return closeNFIFile (file , config->isTesting , config);
}

/********************************************************************/

int main ( int argc, char **argv )
{
    // Parse the command line to get our options and setup
    ConfigInfo *config = openConfigInfo (argc , argv);

    // Start out presuming that we will return success
    // unless we're told otherwise.
    int retVal = EXIT_SUCCESS;

    // If it worked, do something
    if (config != NULL)
    {
        // Did the user ask for help?
        if (config->showHelp)
            printCommandHelp ();

        // Did the user ask for our version?
        else if (config->showVersion)
            printVersion ();

        // Did the user ask to list the contents of a file?
        else if (config->isListing)
            retVal = listContents (config);

        // The only option left is to create/update a file,
        // since the user must specify exactly 1 of list,
        // create or update.
        else 
            retVal = createOrUpdate (config);
    }
    else
    {
        // We had a problem parsing the command line, so that's a failure.
        retVal = 2;
    }

    // Close up the config. This works even if the open function
    // did not.
    closeConfigInfo (config);

    return retVal;
}

/********************************************************************/


