/********************************************************************/

#include "Tool.h"

/********************************************************************/

// This takes a move command and configuration information. If the
// --no-action/-n option is in effect, this will output text to display
// what the command would do.
static void printMoveCommand (CommandStruct *command , ConfigInfo *config)
{
    // If we are not just testing, do nothing here.
    if (config->isTesting == SDL_FALSE)
        return;

    // Based on what move command this is, act accordingly.
    switch (command->cmd)
    {
        case CMD_MOVE:
            printf ("-> %s: All files contained in '%s' whose names match '%s' will be moved to '%s\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->arg , command->toPath);
            break;

        case CMD_MOVERECURSE:
            printf ("-> %s: All files and directories contained in '%s' whose names match '%s' will be moved to '%s\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->arg , command->toPath);
            break;

        case CMD_MOVEFILES:
            printf ("-> %s: Files specified on the command line contained in '%s' will be moved to '%s'\n" ,
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        case CMD_MOVEDIR:
            printf ("-> %s: All files contained in '%s' will be moved to '%s\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        case CMD_MOVEDIRRECURSE:
            printf ("-> %s: All files and directories contained in '%s' will be moved to '%s\n" , 
               optionNameFromValue (command->cmd) , command->fromPath , command->toPath);
            break;

        default:
            break;
    }

    // If this is an option that takes command line files, display them now.
    if (command->cmd == CMD_MOVEFILES)
    {
        printf ("    -> ");

        for (int i = 0 ; i < config->cmdlineFileCount ; i++)
            printf ("%s " , config->cmdlineFiles[i]);

        printf ("\n");
    }
}

/********************************************************************/

// --move: Take a regex and match all filenames in --from to the regex.
// All matched files are moved to the --to path.
//
// --move-recurse: Take a regex and match all filenames in --from and
// it's subdirectories to the regex. All matched files are moved to the --to path.
//
// --move-dir: Find all files whose path is exactly --from
// and move them so that they are in the --to. Unlike --move-dir, this
// only moves files that are directly in the --from path provided.
//
// --move-dir-recurse: Find all files in the NFI whose path contains --from
// and move them so that they exist in --to. 
void cmdMove (NFIFile *nfi , CommandStruct *command , ConfigInfo *config)
{
    // print out what we're going to be doing
    printMoveCommand (command , config);

    // Pretend that we won't recurse with this command.
    SDL_bool recurse = SDL_FALSE;

    // Check to see if the command we're executing is one of the ones that
    // optionally recurse, and set the recurse flag to true if it is.
    if (command->cmd == CMD_MOVERECURSE || command->cmd == CMD_MOVEDIRRECURSE)
        recurse = SDL_TRUE;

    // Pretend that we have no regex
    regex_t *regex = NULL;

    // If this is a command that uses a regex, then set that up
    if (command->cmd == CMD_MOVE || command->cmd == CMD_MOVERECURSE)
        regex = &command->compiledRegex;

    // Now, call into the NFI path structure and get it to move the files for us. If any files
    // moved, then count that as a change to the file
    if (nfi->pathStruct->moveFiles (command->fromPath , command->toPath , regex , recurse , command->clobberIncoming))
        nfi->changeCount++;

    // All done. If we are outputting what we're doing, seperate this command from any that might follow
    if (config->isTesting)
        printf ("\n");
}

/********************************************************************/

// --move-files: Using the filename list specified on the command line,
// move all files who exist in the --from NFI path to the --to NFI path.
void cmdMoveFiles (NFIFile *nfi , CommandStruct *command , ConfigInfo *config)
{
    // print out what we're going to be doing
    printMoveCommand (command , config);

    // Loop over all files that were given on the command line. For each one, construct
    // a full path inside the NFI archive, and then try to move the file from it.
    for (int i = 0 ; i < config->cmdlineFileCount ; i++)
    {
        char thisFile[NFI_NAMELEN + 1];

        // Build a full filename, which consists of the from path, and the path given
        // on the command line (which may also contain path information, which will
        // be considered relative to fromPath)
        sprintf (thisFile , "%s%s" , command->fromPath , config->cmdlineFiles[i]);

        // Try to move the file now. if it worked, count the file as changed.
        if (nfi->pathStruct->moveFile (thisFile , command->toPath , command->clobberIncoming))
            nfi->changeCount++;
    }

    // All done. If we are outputting what we're doing, seperate this command from any that might follow
    if (config->isTesting)
        printf ("\n");
}

/********************************************************************/


