/********************************************************************/

#include "Tool.h"

/********************************************************************/

// Copy dataLength bytes of data from fromHandle to toHandle, using the buffer of the size
// given. The two bools given will be set to FALSE on entry. On exit, if either one is TRUE, it
// tells you what kind of error occured. If both are false, there was no error.
void copyFileData (FILE *fromHandle , FILE *toHandle , Uint32 dataLength , Uint8 *buffer , Uint32 bufferLen , SDL_bool *readError , SDL_bool *writeError)
{
    // Copy all of the bytes over from one handle to the other.
    Uint32 bytesCopied=0,bytesToRead;

    // To begin with, make sure there is no error flagged
    *readError=SDL_FALSE;
    *writeError=SDL_FALSE;

    // Keep looping until we have copied all of the data
    while (bytesCopied != dataLength)
    {
        // Determine how many bytes we want to read this pass.
        bytesToRead = bufferLen;
        if (dataLength - bytesCopied < bytesToRead)
            bytesToRead = dataLength - bytesCopied;

        // Try to read the data, flag an error if we can't read.
        if (fread (buffer , bytesToRead , 1 , fromHandle) != 1)
            *readError = SDL_TRUE;

        // If there wasn't a read error, then try to write.
        if (*readError == SDL_FALSE)
        {
            if (fwrite (buffer , bytesToRead , 1 , toHandle) != 1)
                *writeError = SDL_TRUE;
        }

        // If there was a read or write error, stop copying.
        if (*readError || *writeError)
            break;

        // Update the number of bytes copied.
        bytesCopied += bytesToRead;
    }
}

/********************************************************************/

// Take a mutable path string which may or may not begin with a '/' character. Find
// the next path component and return it. startPos is the position in the string to
// start scanning. The initial call should call this with 0. On successful return,
// it will be set so that successive calls to nextPathPart can be made using the
// value it stores.
//
// Note that the input string will be fiddled with, so it must be an editable
// copy of the string that you will use for no other purpose.
//
// Returns a pointer to a path component on success or NULL if the path is done now.
char *nextPathPart (char *path , int *startPos)
{
    // Alias the scan path to make our lives easier. 
    char *workPath = path + *startPos;
    char *retVal = NULL;

    // Skip past all path seperator characters.
    while (*workPath == '/')
        workPath++;

    // If the next character of the path is a null byte, we're done.
    if (*workPath == '\0')
        return NULL;

    // Save workPath as the value we will return.
    retVal = workPath;

    // Scan workPath forward until it hits a path seperator or the
    // end of the string.
    while (*workPath != '/' && *workPath != '\0')
        workPath++;

    // Here we have found the path seperator or the end of the string.
    // If we found a path seperator, replace it with a null byte and 
    // advance workPath.
    if (*workPath == '/')
    {
        *workPath = '\0';
        workPath++;
    }

    // Set the passed in path to where we stopped for the next call
    *startPos = workPath - path;

    // Return the string we found
    return retVal;
}

/********************************************************************/

// Rebuild the file list in the passed in node (which may be NULL for root) so that any 
// nodes that it is storing that are flagged for deletion will be removed.
void NFIPathStructure::cleanFileList (PathEntry *path)
{
    // Get the list of files from the given node
    PathEntry *fileList = getFileList (path);
    PathEntry *newHead = NULL;

    // Loop over every entry in the given fileList. For each one, 
    // if it is not marked as deleted, link it into the structure
    // at newHead. If it is marked as deleted, then get rid of it.
    while (fileList != NULL)
    {
        // To start with, assume that we are removing no node.
        PathEntry *delNode = NULL;

        // Find and cache the node that is next to this node, since
        // once we perform the relink action below, we will be clobbering
        // it's next during it's reinsertion.
        PathEntry *nextNode = fileList->next;

        // If this node is flagged for deletion, then set delNode
        // to it and do nothing else, so that it will be removed. Otherwise,
        // link it into the list at newHead.
        if (fileList->shouldDelete)
        {
            delNode = fileList;
        } 
        else
        {
            // Link this node in. It's previous should be nothing, because
            // it will be the new head. It's next should be whatever the
            // current head it.
            fileList->prev = NULL;
            fileList->next = newHead;

            // If there is currently an entry at the list head, it's prev
            // should point to this new node.
            if (newHead != NULL)
                newHead->prev = fileList;

            // Now, this node is the new head.
            newHead = fileList;
        }

        // Go to the next node now. We saved the next at the start of this
        // iteration because the relink operation above clobbers out the
        // next.
        fileList = nextNode;

        // If delNode is a valid node, this node was flagged for deletion,
        // so get rid of it.
        if (delNode != NULL)
        {
            free (delNode->name);
            free (delNode);
        }
    }

    // Now, replace the file list in the given path with the list at
    // newHead. Here we can't use accessors and have to do the job
    // manually. This might end up as a NULL list, if all files were
    // flagged for removal.
    if (path == NULL)
        files = newHead;
    else
        path->files = newHead;
}

/********************************************************************/

// Given a path entry (which may be NULL for the root of the path system), return back a pointer to the
// file or directory list of that node. These helpers exist because the data for the root appears in
// members of the class, but the data for other paths appear in the nodes themselves.
PathEntry *NFIPathStructure::getFileList (PathEntry *path)
{
    // Return the member variable for the ROOT path.
    if (path == NULL)
        return files;
    else
        return path->files;
}

/********************************************************************/

// Given a path entry (which may be NULL for the root of the path system), return back a pointer to the
// file or directory list of that node. These helpers exist because the data for the root appears in
// members of the class, but the data for other paths appear in the nodes themselves.
PathEntry *NFIPathStructure::getDirectoryList (PathEntry *path)
{
    // Return the member variable for the ROOT path.
    if (path == NULL)
        return directories;
    else
        return path->directories;
}

/********************************************************************/

// Return a full path from root to the passed in node in a static
// char buffer. This makes it quite, quite unthreadsafe. If the node
// passed in is NULL, the string "/" is returned.
const char *NFIPathStructure::pathFromNode (PathEntry *node)
{
    static char buffer[1024];

    // If the node we were given isn't NULL, call ourselves again with 
    // this node's parent node. When this doesn't happen, node is finally
    // null, which means we've hit the top of the directory tree.
    if (node != NULL)
        pathFromNode (node->parent);

    // We hit here first when node is NULL, then once for every node going
    // back to where we started as each recursive invocation of the function
    // terminates. If the node is NULL, then this is the root, so take that
    // opportunity to start the string. Otherwise, we should append the
    // path of the current node to the string.
    if (node == NULL)
        strcpy (buffer , "/");
    else
    {
        // Not the root, so append the path of this node and a seperator.
        strcat (buffer , node->name);
        strcat (buffer , "/");
    }

    // Return the buffer as it stands now. WHen the original invocation
    // exist, the buffer is a fully formed path.
    return buffer;
}

/********************************************************************/

// Construct a new structure based off of a preloaded file.
NFIPathStructure::NFIPathStructure (NFIFile *file , SDL_bool verbose)
{
    // Start empty
    directories = NULL;
    files = NULL;
    extractBuffer = NULL;

    // Scan the provided file
    bVerbose = SDL_FALSE;
    scanFile (file);

    // Now, set the verbose flag. We do this now because if we are
    // verbose while scanFile runs, it will announce that it has added
    // files while it's scanning the archive and building the initial
    // structure.
    bVerbose = verbose;

    // Save a pointer to the archive.
    archive = file;
}

/********************************************************************/

// Takes a path node that is the head of a list of files, and removes every file in the list.
void NFIPathStructure::removeFileList (PathEntry *head)
{
    PathEntry *tmp;

    // Loop over all files.
    while (head != NULL)
    {
        // Store this file then advance to the next.
        tmp = head;
        head = head->next;

        // Delete the head now.
        free (tmp->name);
        free (tmp);
    }
}

/********************************************************************/

// Takes a path node that is the head of a list of directories. All directories and subdirectories
// of the given node are removed, as well as any files they are storing.
void NFIPathStructure::removeDirList (PathEntry *head)
{
    PathEntry *tmp;

    // Loop over all directories
    while (head != NULL)
    {
        // Remove all files in this directory.
        removeFileList (head->files);

        // Now, recursively call ourselves for our directory list, to make sure
        // everything under us is gone.
        removeDirList (head->directories);

        // Save this node and advance to the next.
        tmp = head;
        head = head->next;

        // Now we can release this node.
        free (tmp->name);
        free (tmp);
    }
}

/********************************************************************/

NFIPathStructure::~NFIPathStructure ()
{
    // Remove the files and directories stored in the root, and by
    // definition every directory.
    removeFileList (files);
    removeDirList (directories);

    // If there is an extraction buffer, get rid of it
    if (extractBuffer != NULL)
        free (extractBuffer);
}

/********************************************************************/

// Display a node in the tree.
void NFIPathStructure::printNode (PathEntry *node , int indent)
{
    // Start assuming that the file and directory nodes are for the
    // root path.
    PathEntry *fileNode = files;
    PathEntry *directoryNode = directories;
    PathEntry *tmp;
    const char *pathName = "";

    // If the node given to print isn't NULL, then get it's name
    // and nodes for display.
    if (node != NULL)
    {
        pathName      = node->name;
        fileNode      = node->files;
        directoryNode = node->directories;
    }

    // Build an indent string.
    char indentStr[1024]="";
    for (int ind = 0 ; ind < indent ; ind++)
        strcat (indentStr , "|  ");

    // Display the path name
    printf (" %s%s/\n" , indentStr , pathName);

    // Now recursively call ourselves for each path
    // stored under us.
    tmp = directoryNode;
    while (tmp != NULL)
    {
        printNode (tmp , indent + 1);
        tmp = tmp->next;
    }

    // Display all files we contain
    tmp = fileNode;
    while (tmp != NULL)
    {
        printf (" %s `- %s" , indentStr , tmp->name);

        switch (tmp->info->type)
        {
            case NFI_UNTOUCHED:    printf ("\n"); break;
            case NFI_ADDED_FILE:   printf (" (+)\n"); break;
            case NFI_REMOVED_FILE: printf (" (-)\n"); break;
            case NFI_MOVED_FILE:   printf (" (~)\n"); break;
        }

        tmp = tmp->next;
    }
}

/********************************************************************/

// Output the contents of the NFI structure to stdout
void NFIPathStructure::displayContents ()
{
    printf ("\n---------------------------------------\n\n");

    // Print the root path and recursively, all files under it
    printNode (NULL , 0);

    printf ("\n---------------------------------------\n\n");
}

/********************************************************************/

// Scan the provided NFIFile structure and add all files and paths in it to ourselves.
void NFIPathStructure::scanFile (NFIFile *file)
{
    // Scan over all directory entries in the given file. For each one,
    // add in the file.
    for (Uint32 i = 0 ; i < file->dirHeader.dirSize ; i++)
    {
        // Look up the path of this file
        PathEntry *path = findPath (file->directory[i].path);

        // Add it in
        addFile (path , file->directory[i].name , &(file->directory[i]) , SDL_TRUE);
    }
}
       
/********************************************************************/

// Add a new node to the given node. if isFile is TRUE, the new node is added to the file
// list of the provided node, otherwise it is added to the directory list of the added
// node. In either case, the newly inserted node is returned back to you. The new node will
// be fully initialized, though if it is a file node it is up to the caller to add in the
// NFI info.
PathEntry *NFIPathStructure::addNodeToPath (PathEntry *node , const char *name , SDL_bool isFile)
{
    // Allocate and initialize a new node.
    PathEntry *newNode = (PathEntry *) malloc (sizeof (PathEntry));
    memset (newNode , 0 , sizeof (PathEntry));

    // Copy over the name, and save the parent
    newNode->name   = strdup (name);
    newNode->parent = node;
    
    // This is the node that we want to insert into. This is an alias here
    // even though we essentially get the value we want passed in, because
    // if the value passed in is NULL, we will modify the class variables
    // (they represent the rool path, and so does NULL in this context).
    // Additionally, we might chose a subnode of the node passed in to add, 
    // based on the value of isFile.
    //
    // Note that this is a pointer to a pointer to a node, so that we can
    // modify the contents indirectly.
    PathEntry **parentNode = NULL;
   
    // If the node to add to is NULL, the node should be added to the root
    // dir, which is the list variables directly in the class.
    if (node == NULL)
    {
        // We are adding something directly to the root path. This is represented
        // by our instance variables "directories" if we are adding a directory,
        // or "files" if we are adding a file. 
        if (isFile)
            parentNode = &files;
        else
            parentNode = &directories;
    }
    else
    {
        // The node provided is valid, so we will add this entry to it. If isFile
        // is TRUE, we are added to the files member of the node. Otherwise, we 
        // link ourselves into the node provided.
        if (isFile)
            parentNode = &(node->files);
        else
            parentNode = &(node->directories);
    }

    // If the parent node is NULL, then our job is easy, because we just set it equal
    // to the new node and that's that.
    if (*parentNode == NULL)
        *parentNode = newNode;
    else
    {
        // The parent node is not null. In that case, the next member of the new node must
        // point to the parent.
        newNode->next = *parentNode;

        // Next, the previous member of the parent must point to the new node.
        (*parentNode)->prev = newNode;

        // Now, make the parent node be the new node.
        *parentNode = newNode;
    }

    // Return the newly allocated and inserted node, so that if it is a file,
    // extra operations can be done on it.
    return newNode;
}

/********************************************************************/

// Search the given node to see if it contains the name provided. If it
// does, return a pointer to the node containing that name. Otherwise,
// return NULL. NULL is returned if name is "/" or "" as well, since that
// is technically the node that houses the root dir.
//
// If isFile is true, the file list in the passed in node is searched. 
// Otherwise, the directory list is searched.
//
// Note that the passed in node can be NULL (root or /).
PathEntry *NFIPathStructure::findEntryNamed (PathEntry *node , const char *name , SDL_bool isFile)
{
    PathEntry *searchPos;

    // Get the list of files and directories to search.
    if (isFile)
        searchPos = getFileList (node);
    else
        searchPos = getDirectoryList (node);

    // While the search position isn't NULL, try to find a node
    // that matches.
    while (searchPos != NULL)
    {
        // If this node matches, return it.
        if (strcmp (searchPos->name , name) == 0)
            return searchPos; 

        // Go to the next entry now
        searchPos = searchPos->next;
    }

    // If we get here, no match.
    return NULL;
}

/********************************************************************/

// Find and return a pointer to a given path. If the path
// given doesn't exist yet in the path structure, it will
// be created for you.
PathEntry *NFIPathStructure::findPath (const char *path)
{
    // This is our return value. It starts out as NULL (root path (/))
    // and as we scan path components it is set to their path entries
    // in the structure. Path component 1 is searched for relative to 0,
    // 2 is relative to 1, and so on. When all components have been
    // scanned, this is the node that we want.
    //
    // During scanning if a target node doesn't have a required child
    // node, it will be added.
    PathEntry *retVal = NULL;

    // Generate a temporary copy of the path provided, so that
    // we can fiddle with it.
    char tempPath[NFI_NAMELEN + 1];
    strcpy (tempPath , path);

    // Start at the start of the passed in path.
    int pos=0;
    char *part;

    // Find the first part of the path.
    part = nextPathPart (tempPath , &pos);

    // While we have a path part, work on it.
    while (part != NULL)
    {
        // Try to find a directory named part in the current 
        // node.
        PathEntry *found = findEntryNamed (retVal , part , SDL_FALSE);

        // If the entry was not found, add it.
        if (found == NULL)
            found = addNodeToPath (retVal , part , SDL_FALSE);

        // Now, whatever node we found/created is the new current node
        retVal = found;

        // Step to the next part of the path now
        part = nextPathPart (tempPath , &pos);
    }

    // Return the resulting path component, which might be NULL if the path passed
    // in was "/" or "".
    return retVal;
}

/********************************************************************/

// Add a file at a given predefined path or by parsing the path
// given. When clobber is TRUE, if there is a file with the name
// given at the path given, it will be replaced with the new file
// (that is, the NFI info will be copied overtop of the old file)
// Otherwise, such a file will not be touched.
SDL_bool NFIPathStructure::addFile (PathEntry *path , char *filename , NFIParsedDirEntry *info , SDL_bool clobber)
{
    // Search to see if there is an existing entry at this path with this name.
    PathEntry *node = findEntryNamed (path , filename , SDL_TRUE);

    // If clobber is set to false and we found an entry and the entry is not marked
    // as removed, then don't add
    if (clobber == SDL_FALSE && node != NULL && node->info->type != NFI_REMOVED_FILE)
    {
        if (bVerbose)
            fprintf (stderr , "  -> Protected: %s%s\n", node->info->path , filename);
        return SDL_FALSE;
    }

    // Here we want to clobber a file if it exists or add it if it doesn't. If we have an
    // existing node, then replace it's name and info with what was passed in.
    if (node)
    {
        // Replace the name. We leave everything else as is, since the node is already
        // linked into the system properly.
        free (node->name);
        node->name = strdup (filename);

        // Flag the info in this node as the file being removed, because it is being
        // replaced with a new entry added elsewhere.
        node->info->type = NFI_REMOVED_FILE;
    }
    else
    {
        // There was no existing node, so add one now.
        node = addNodeToPath (path , filename , SDL_TRUE);
    }

    // Set the info into the node
    node->info = info;

    // Tell the info in the node who owns it. When a file add operation begins,
    // we realloc the directory structure of the loaded NFI file. At that point,
    // all existing PathEntry pointers in the structure tree become invalid,
    // as their info structure is pointing to the memory where the files used to
    // be instead of where they are now (because realloc may potentially return
    // a different block of memory entirely). By telling each info where it's
    // node structure is, after a realloc we can come back to these nodes and
    // update them properly.
    node->info->entryPtr = node;

    // If we're verbose, say we added the file
    if (bVerbose)
        printf ("  -> Added:     %s%s\n" , info->path , info->name);

    return SDL_TRUE;
}

/********************************************************************/

// Add based on a pre-filled out info structure.
SDL_bool NFIPathStructure::addFile (NFIParsedDirEntry *info , SDL_bool clobber)
{
    PathEntry *rPath = findPath (info->path);

    return addFile (rPath , info->name , info , clobber);
}

/********************************************************************/

// Flag the file with the given name (including path), if it exists.
SDL_bool NFIPathStructure::removeFile (const char *name)
{
    char path[NFI_NAMELEN + 1];
    char *part;

    // Make a copy of the name given in.
    strcpy (path , name);

    // Find the last path seperator in the string. If there 
    // wasn't one, we were called wrong so return.
    part = strrchr (path , '/');
    if (part == NULL)
        return SDL_FALSE;

    // alias the name to be the character after the path
    // sep, then blow the path sep away.
    name = part + 1;
    *part = '\0';

    // Now, find the path
    PathEntry *rPath = findPath (path);

    // Check the files in this path (which might be NULL), to see
    // if the file with the given name exists or not.
    PathEntry *fileList = getFileList (rPath);

    // Scan the files. When we find the match we want, flag it as
    // removed and return.
    while (fileList != NULL)
    {
        // If this is the file we want to delete, flag it to be gone
        // and then leave.
        if (strcmp (name , fileList->name) == 0)
        {
            // If we're verbose, say we removed the file
            if (bVerbose)
                printf ("  -> Removed:   %s%s\n" , fileList->info->path , fileList->info->name);

            fileList->info->type = NFI_REMOVED_FILE;
            return SDL_TRUE;
        }

        // Next
        fileList = fileList->next;
    }

    // No file removed
    return SDL_FALSE;
}

/********************************************************************/

// Move the file with the given name (including path) to the other path,
// if it exists
SDL_bool NFIPathStructure::moveFile (const char *name , const char *destPath , SDL_bool clobber)
{
    char path[NFI_NAMELEN + 1];
    char *part;

    // Make a copy of the name given in.
    strcpy (path , name);

    // Find the last path seperator in the string. If there 
    // wasn't one, we were called wrong so return.
    part = strrchr (path , '/');
    if (part == NULL)
        return SDL_FALSE;

    // alias the name to be the character after the path
    // sep, then blow the path sep away.
    name = part + 1;
    *part = '\0';

    // Now, find the paths
    PathEntry *fromPath = findPath (path);
    PathEntry *toPath = findPath (destPath);

    // Check the files in this path (which might be NULL), to see
    // if the file with the given name exists or not.
    PathEntry *fileList = getFileList (fromPath);

    // Scan the files. 
    while (fileList != NULL)
    {
        // We have work to do if this is the file we want to move.
        if (strcmp (name , fileList->name) == 0)
        {
            // Try to move the file, then break out of the loop.
            performMove (fromPath , toPath , fileList , NULL , clobber);
            break;
        }

        // Next
        fileList = fileList->next;
    }

    // At this point, if fileList is NULL, then we scanned all files and
    // moved none because none matched. If however we did move a file, then
    // the list broke leaving fileList as the node that was moved. In that
    // case, we need to do some cleanup.
    if (fileList != NULL)
    {
        cleanFileList (fromPath);
        return SDL_TRUE;
    }

    // Nothing was moved.
    return SDL_FALSE;
}

/********************************************************************/

// Flag files at the given path node as removed. If recursive is true, also
// flag files in every node underneath this one.
SDL_bool NFIPathStructure::removeFiles (PathEntry *path , regex_t *regex , SDL_bool recursive)
{
    int removedCount=0;

    // If the root path given was NULL, then the files and directories to scan are
    // the members of the class. Otherwise, they're from the path.
    PathEntry *fileList = getFileList (path);
    PathEntry *dirList  = getDirectoryList (path);

    // To start with, blow away all files in the provided path.
    while (fileList)
    {
        // If there is no regex, or there is and it matches this name,
        // then flag it as removed.
        if (regex == NULL || (regex != NULL && regexec (regex , fileList->name , 0 , 0 , 0) == 0))
        {
            // If we're verbose, say we removed the file
            if (bVerbose)
                printf ("  -> Removed:   %s%s\n" , fileList->info->path , fileList->info->name);

            // Flag the info in the file as removed
            fileList->info->type = NFI_REMOVED_FILE;

            // Count this as a file removed
            removedCount++;
        }

        // Next file
        fileList = fileList->next;
    }

    // If we are recursive, then call ourselves for all of our paths
    // as well.
    if (recursive)
    {
        // Call ourselves for all paths we store.
        while (dirList)
        {
            // Try to remove files from our subdir. If it removed any files, then
            // count that as a removal for here. This makes sure that even if we
            // removed no files ourself, if some file under us was removed, this
            // will flag it.
            if (removeFiles (dirList , regex , SDL_TRUE) == SDL_TRUE)
                removedCount++;

            // Next dir
            dirList = dirList->next;
        }
    }

    // If some files were moved, return success
    if (removedCount != 0)
        return SDL_TRUE;

    // No files were removed
    return SDL_FALSE;
}

/********************************************************************/

// Flag files at the given path node as removed. If recursive is true, also
// flag files in every node underneath this one.
SDL_bool NFIPathStructure::removeFiles (const char *path , regex_t *regex , SDL_bool recursive)
{
    // Find the path then call the other functions
    PathEntry *rPath = findPath (path);

    return removeFiles (rPath , regex , recursive);
}

/********************************************************************/

// Perform an actual move operation. The file is moved from the path specified
// by fromPath to the path specified in toPath, and it is assumed that
// fileNode is a valid node in the path that is stored in fromPath. The file
// will only be moved if:
//   a) it is not flagged as removed
//   b) it matches the regex given (all files match if regex is NULL)
//   c) there is no file with the same name in the toPath, or there is but clobber is TRUE.
//
// If a file is moved, the node it came from (under fromPath) is flagged for removal. The
// info structure in the file will be modified to have the proper path. if the file moved
// was an added file, nothing else happens. otherwise, the file is flagged as moved.
//
// The return value tells you if the file was moved or not
SDL_bool NFIPathStructure::performMove (PathEntry *fromPath , PathEntry *toPath , PathEntry *fileNode , regex_t *regex , SDL_bool clobber)
{
    // If this file is flagged as being a removed file, we do not consider it for the move, because
    // there is no point.
    if (fileNode->info->type == NFI_REMOVED_FILE)
        return SDL_FALSE;

    // If there is a regex and this name doesn't match it, then return because we don't want to move
    // the file.
    if (regex != NULL && regexec (regex , fileNode->name , 0 , 0 , 0) == REG_NOMATCH)
        return SDL_FALSE;

    // Scan the "to" path to see if there is an entry at that path that exists already.
    PathEntry *node = findEntryNamed (toPath , fileNode->name, SDL_TRUE);

    // If we found a node, we have to check to see if we should move the file or not, since
    // by doing so we will clobber over a node. We don't want to clobber if clobber is set
    // to FALSE. However, if the file is marked as removed, then it is ok to clobber it no
    if (clobber == SDL_FALSE && node != NULL && node->info->type != NFI_REMOVED_FILE)
    {
        // We will not move because we would be clobbering a file. If we are in verbose
        // mode, then tell someone that the file would be protected
        if (bVerbose)
            fprintf (stderr , "  -> Protected: %s%s\n", node->info->path , fileNode->name);

        // Done now
        return SDL_FALSE;
    }

    // If the code above found a node that exists in the "to" path whose name matches
    // the file we are moving, and we get here, then we must be clobbering that
    // data. In tha case, we should flag the info in that node as removed, because it
    // is effectively gone now.
    //
    // If there wasn't a node found, then we don't have to do anything at this point
    // but allocate a new node in the "to" path to store the file being moved.
    if (node != NULL)
    {
        // node points to a file being stored in toPath. Flag the info in the node as 
        // being removed, because it is going to be superceded by an existing file.
        node->info->type = NFI_REMOVED_FILE;
    }
    else
    {
        // There was no matching file in the toPath, so add a node there now.
        node = addNodeToPath (toPath , fileNode->name , SDL_TRUE);
    }

    // Now, the info structure in fileNode is the structure in the fromPath that
    // represents this file. Shift it's info structure to point to the node in toPath
    // (represented by node), because that is where the file lives now.
    node->info = fileNode->info;

    // Tell the info structure in the node that it's now being stored in this node
    // instead of the other one. Otherwise if there is another addition and the
    // pointers need to readjust, it will try to adjust a non-existant node.
    node->info->entryPtr = node;

    // If the file wasn't added, then flag it as moved. If it was flagged as added,
    // then we will leave the flag alone, though the path will get modified below.
    // This makes sure that files get added to the right place, but that if they
    // are newly added and then moved, they are just considered added at a different
    // path.
    if (node->info->type != NFI_ADDED_FILE)
        node->info->type = NFI_MOVED_FILE;

    // Copy into the node the path of the current toPath, so that the file tracks
    // it's position properly.
    strcpy (node->info->path , pathFromNode (toPath));

    // If we're verbose, say what we did
    if (bVerbose)
        printf ("  -> Moved:     %s%s to %s\n" , pathFromNode (fromPath) , node->info->name , node->info->path);

    // The node currently being pointed at by fileNode is now a node in the fromPath that represents a file
    // that we no longer want. In this case, flag this node for deletion once the move operation completes.
    fileNode->shouldDelete = SDL_TRUE;

    // Move worked
    return SDL_TRUE;
}

/********************************************************************/

// Move all files from the given fromPath node to the toPath node. If recursive is
// TRUE, then directory nodes are also moved.
//
// All affected files have their NFI info updated to reflect their new path.
//
// If clobber is false, any file in fromPath that already exists in toPath 
// will not be moved (that is, it's NFI info won't be updated and the path
// it was stored in won't be fully moved). Otherwise, everything is moved
// wholesale to the toPath.
//
// The move will skip files marked as removed. If a file is marked as added, then it's internal
// path and location in the path tree will be altered, but the file itself will remain flagged
// as added. Otherwise, the file will be flagged as moved. 
SDL_bool NFIPathStructure::moveFiles (PathEntry *fromPath , PathEntry *toPath , regex_t *regex , SDL_bool recursive, SDL_bool clobber)
{
    // If the from location is the NULL path, then the files and directories that we 
    // want to move (or consider moving) are the members of the class. otherwise, 
    // they're from the path provided.
    PathEntry *fileList = getFileList (fromPath);
    PathEntry *dirList  = getDirectoryList (fromPath);
    int movedFiles=0;

    // To start with, move all files contained in the from path to the to path.
    while (fileList)
    {
        // Try to actually move this file. If it worked, increment the count of
        // files that were moved.
        if (performMove (fromPath , toPath , fileList , regex , clobber))
            movedFiles++;

        // Go no to the next file
        fileList = fileList->next;
    }

    // For every file that was moved there now exists a node in fromPath->files
    // that is marked for deletion. If that is the case, we should rebuild the
    // file list in fromPath to fix the problem.
    if (movedFiles != 0)
        cleanFileList (fromPath);

    // If we are recursive, then call ourselves for all of our paths
    // as well.
    if (recursive)
    {
        // Call ourselves for all paths we store.
        while (dirList)
        {
            // First of all, for the name of this directory, find that
            // directory in the to path. If it doesn't exist, then add it.
            // use this as the new "to" directory.
            PathEntry *newDest = findEntryNamed (toPath , dirList->name , SDL_FALSE);
            
            // Didn't find a proper directory in the to path, so add it
            if (newDest == NULL)
                newDest = addNodeToPath (toPath , dirList->name , SDL_FALSE);

            // Call ourselves now. If this returns true, then at least one file
            // at this path or below was moved, so count that as a move here
            if (moveFiles (dirList , newDest , regex , recursive , clobber))
                movedFiles++;

            // Next item
            dirList = dirList->next;
        }
    }

    // If any files were moved, return success.
    if (movedFiles != 0)
        return SDL_TRUE;

    // No files moved
    return SDL_FALSE;
}

/********************************************************************/

// Move all files from the given fromPath node to the toPath node. If recursive is
// TRUE, then directory nodes are also moved.
//
// All affected files have their NFI info updated to reflect their new path.
//
// If clobber is false, any file in fromPath that already exists in toPath 
// will not be moved (that is, it's NFI info won't be updated and the path
// it was stored in won't be fully moved). Otherwise, everything is moved
// wholesale to the toPath.
SDL_bool NFIPathStructure::moveFiles (const char *fromPath , const char *toPath , regex_t *regex , SDL_bool recursive, SDL_bool clobber)
{
    PathEntry *from = findPath (fromPath);
    PathEntry *to   = findPath (toPath);

    return moveFiles (from , to , regex , recursive , clobber);
}

/********************************************************************/

// Perform an actual extraction.
SDL_bool NFIPathStructure::performExtract (PathEntry *fileNode , const char *toPath , regex_t *regex , SDL_bool clobber)
{
    // If this file is flagged as being a removed file, we do not consider it for the extraction, because
    // there is no point.
    if (fileNode->info->type == NFI_REMOVED_FILE)
        return SDL_FALSE;

    // If there is a regex and this name doesn't match it, then return because we don't want to extract
    // the file.
    if (regex != NULL && regexec (regex , fileNode->name , 0 , 0 , 0) == REG_NOMATCH)
        return SDL_FALSE;

    // Check to see if there is a file at toPath with the name we have here, and if there is, and clobber
    // is FALSE, do not clobber the file.
    char outName[NFI_NAMELEN + 1];
    sprintf (outName , "%s/%s" , toPath , fileNode->name);

    // If a file with this name exists and clobber is FALSE, then we would be overwriting a file, so 
    // don't do that.
    if (realFileExists (outName) == SDL_TRUE && clobber == SDL_FALSE)
    {
        if (bVerbose)
            fprintf (stderr , "  -> Protected: %s\n", outName);

        return SDL_FALSE;
    }

    // We are going to overwrite the given file, or it doesn't exist. To start with, open the
    // output file for writing.
    FILE *inHandle,*outHandle;
    Sint32 fileLength;

    // We have determined that either the file doesn't exist, or it does and we want to clobber
    // it. Ensure that the full extraction path up to this file exists.
    createFullPath (outName);

    // Try to open the file. This might not work if permissions are wrong, the create path
    // function didn't work, or we couldn't open the file because of permissions (in which 
    // case we can't write it either, probably).
    outHandle = fopen (outName , "wb");
    if (outHandle == NULL)
    {
        fprintf (stderr , "Unable to open '%s' for writing\n" , outName);
        return SDL_FALSE;
    }

    // Get a handle for output. If the file we are extracting is flagged as "added", then
    // we need to open a file. Otherwise, we can just use the one from the archive.
    if (fileNode->info->type == NFI_ADDED_FILE)
    {
        // Try to open the file. This may or may not work.
        inHandle = fopen (fileNode->info->sourceName , "rb");
        if (inHandle == NULL)
        {
            // Close up and get rid of the output file.
            fclose (outHandle);
            unlink (outName);

            // Error out.
            fprintf (stderr , "Unable to open '%s' for reading\n" , fileNode->info->sourceName);
            return SDL_FALSE;
        }

        // Find out the length of the file
        fseek (inHandle , 0 , SEEK_END);
        fileLength = ftell (inHandle);
        fseek (inHandle , 0 , SEEK_SET);
    }
    else
    {
        // The file isn't being added, so it already exists in the archive. In this case,
        // set up the handle from the archive and seek to the start of the file.
        inHandle = archive->fileHandle;
        fseek (inHandle , fileNode->info->offset , SEEK_SET);
        fileLength = fileNode->info->length;
    }

    // We're extracting something and have the handles we need. If we're verbose, say
    // what we're doing.
    if (bVerbose)
        fprintf (stderr , "  -> Extracted: %s\n", outName);

    // Allocate a buffer if we don't already have one
    if (extractBuffer == NULL)
        extractBuffer = (Uint8 *) malloc (EXTRACT_BUF_BYTES);
    
    // Copy all of the bytes over from one handle to the other.
    SDL_bool readError,writeError;
    copyFileData (inHandle , outHandle , fileLength , extractBuffer , EXTRACT_BUF_BYTES , &readError , &writeError);

    // Here the file is either copied, or there was an error. To start with,
    // if we opened a handle for reading, close it. We leave it alone if the
    // file we were reading from is the archive.
    if (fileNode->info->type == NFI_ADDED_FILE)
        fclose (inHandle);

    // Close the output handle.
    fclose (outHandle);

    // If there was an error, then we want to unlink the output file, and then
    // complain and return failure.
    if (readError || writeError)
    {
        // Get rid of the partially extracted file.
        unlink (outName);

        // Display an error and return failure
        fprintf (stderr , "%s error extracting file %s\n" , (readError == SDL_TRUE ? "Read" : "Write") , outName);
        return SDL_FALSE;
    }

    // Worked
    return SDL_TRUE;
}

/********************************************************************/

// Move files from the fromPath in the archive to the toPath in the local filesystem. 
void NFIPathStructure::extractFiles (PathEntry *fromPath , const char *toPath , regex_t *regex , SDL_bool recursive, SDL_bool clobber)
{
    // If the from location is the NULL path, then the files and directories that we 
    // want to extract (or consider extracting) are the members of the class. otherwise, 
    // they're from the path provided.
    PathEntry *fileList = getFileList (fromPath);
    PathEntry *dirList  = getDirectoryList (fromPath);

    // To start with, extract all files contained in the from path to the to path.
    while (fileList)
    {
        // Try to extract this file
        performExtract (fileList , toPath , regex , clobber);

        // Go no to the next file
        fileList = fileList->next;
    }

    // If we are recursive, then call ourselves for all of our paths
    // as well.
    if (recursive)
    {
        // Call ourselves for all paths we store.
        while (dirList)
        {
            char newToPath[NFI_NAMELEN + 1];

            // First of all, construct a new toPath that consists of the toPath that we
            // were given, plus the name of the path in this dir.
            sprintf (newToPath , "%s/%s" , toPath , dirList->name);

            // Now recursively call ourselves to extract files in this path.
            extractFiles (dirList , newToPath , regex , recursive , clobber);

            // Next item
            dirList = dirList->next;
        }
    }
}

/********************************************************************/

// Move files from the fromPath in the archive to the toPath in the local filesystem. 
void NFIPathStructure::extractFiles (const char *fromPath , const char *toPath , regex_t *regex , SDL_bool recursive , SDL_bool clobber)
{
    PathEntry *rPath = findPath (fromPath);

    extractFiles (rPath , toPath , regex , recursive , clobber);
}

/********************************************************************/

// Extract a single file to the given to path, optionally not clobbering the file.
void NFIPathStructure::extractFile (const char *name , const char *toPath , SDL_bool clobber)
{
    char path[NFI_NAMELEN + 1];
    char *part;

    // Make a copy of the name given in.
    strcpy (path , name);

    // Find the last path seperator in the string. If there 
    // wasn't one, we were called wrong so return.
    part = strrchr (path , '/');
    if (part == NULL)
        return;

    // alias the name to be the character after the path
    // sep, then blow the path sep away.
    name = part + 1;
    *part = '\0';

    // Now, find the path
    PathEntry *fromPath = findPath (path);

    // Check the files in this path (which might be NULL), to see
    // if the file with the given name exists or not.
    PathEntry *fileList = getFileList (fromPath);

    // Scan the files. 
    while (fileList != NULL)
    {
        // We have work to do if this is the file we want to move.
        if (strcmp (name , fileList->name) == 0)
        {
            // Try to move the file, then return because we're done
            performExtract (fileList , toPath , NULL , clobber);
            return;
        }

        // Next
        fileList = fileList->next;
    }
}

/********************************************************************/

// Take a path that is either relative to the current directory or
// an absolute path. Make sure that all paths along this path
// exist. The name passed in may optionally contain a filename spec.
// If it doesn't, the name must terminate in a '/' character.
void NFIPathStructure::createFullPath (const char *name)
{
    char path[NFI_NAMELEN + 1];
    char cmd[NFI_NAMELEN +25];
    char *pathPart;

    // Make a copy of what was passed in
    strcpy (path , name);

    // Try to find the last path seperator. If we find one, clobber
    // it out.
    pathPart = strrchr (path , '/');
    if (pathPart != NULL)
        *pathPart = '\0';

    // Now, build a command line
    sprintf (cmd , "mkdir -p %s" , path);

    // Execute it
    system (cmd);
}

/********************************************************************/

// Check to see if there is a file at the path provided. If there is, return TRUE. Otherwise,
// return false. This checks the local filesystem. The name passed can therefore be absolute,
// or relative to the current directory, as needed.
SDL_bool NFIPathStructure::realFileExists (const char *name)
{
    // Try to open the file
    FILE *f = fopen (name , "rb");

    // If the open failed, pretend the file didn't exist.
    if (f == NULL)
        return SDL_FALSE;

    // We could open the file, so close it and return that it exists.
    fclose (f);
    return SDL_TRUE;
}

/********************************************************************/

