/********************************************************************/

#include "Tool.h"

/********************************************************************/

// This takes a remove command and configuration information. If the
// --no-action/-n option is in effect, this will output text to display
// what the command would do.
static void printRemoveCommand (CommandStruct *command , ConfigInfo *config)
{
    // If we are not just testing, do nothing here.
    if (config->isTesting == SDL_FALSE)
        return;

    // Based on what remove command this is, act accordingly.
    switch (command->cmd)
    {
        case CMD_REMOVE:
            printf ("-> %s: Files in NFI path '%s' whose names match '%s' will be removed from the NFI archive\n" ,
               optionNameFromValue (command->cmd) , command->fromPath , command->arg);
            break;

        case CMD_REMOVERECURSE:
            printf ("-> %s: Files in NFI path '%s' and it's children whose names match '%s' will be removed from the NFI archive\n" ,
               optionNameFromValue (command->cmd) , command->fromPath , command->arg);
            break;

        case CMD_REMOVEFILES:
            printf ("-> %s: Files from the command line will be removed from the NFI if they exist in the NFI path '%s'\n" ,
               optionNameFromValue (command->cmd) , command->fromPath);
            break;

        case CMD_REMOVEDIR:
            printf ("-> %s: All files in NFI path '%s' will be removed from the NFI archive\n" ,
               optionNameFromValue (command->cmd) , command->fromPath);
            break;

        case CMD_REMOVEDIRRECURSE:
            printf ("-> %s: All files and subdirectories of NFI path '%s' will be removed from the NFI archive\n" ,
               optionNameFromValue (command->cmd) , command->fromPath);
            break;

        default:
            break;
    }

    // If this is an option that takes command line files, display them now.
    if (command->cmd == CMD_REMOVEFILES)
    {
        printf ("    -> ");

        for (int i = 0 ; i < config->cmdlineFileCount ; i++)
            printf ("%s " , config->cmdlineFiles[i]);

        printf ("\n");
    }
}

/********************************************************************/

// --remove: Argument is a regex. Scan all files in the --from NFI
// path. If the filename portion of files that fall exactly in the
// --from path match the regex, they are removed.
//
// --remove-recurse: Argument is a regex.Scan all files in the --from
// NFI path, and all filenames that are under that path. If the filename
// portion of the files found match the regex, they are removed.
//
// --remove-dir: Remove all files that fall exactly in the --from NFI
// dir. Files in subdirectories of this path are left alone.
//
// --remove-dir-recurse: Scan for all files in the NFI that exist in the --from
// path, including files that are in subdirectories of that path. Any
// files thus found are removed from the file.
void cmdRemove (NFIFile *nfi , CommandStruct *command , ConfigInfo *config)
{
    // print out what we're going to be doing
    printRemoveCommand (command , config);

    // Pretend that we won't recurse with this command.
    SDL_bool recurse = SDL_FALSE;

    // Check to see if the command we're executing is one of the ones that
    // optionally recurse, and set the recurse flag to true if it is.
    if (command->cmd == CMD_REMOVERECURSE || command->cmd == CMD_REMOVEDIRRECURSE)
        recurse = SDL_TRUE;

    // Pretend that we have no regex
    regex_t *regex = NULL;

    // If this is a command that uses a regex, then set that up
    if (command->cmd == CMD_REMOVE || command->cmd == CMD_REMOVERECURSE)
        regex = &command->compiledRegex;

    // Now, call into the NFI path structure and get it to remove files for us. If any
    // files were removed, count the file as changed.
    if (nfi->pathStruct->removeFiles (command->fromPath , regex , recurse))
            nfi->changeCount++;

    // All done. If we are outputting what we're doing, seperate this command from any that might follow
    if (config->isTesting)
        printf ("\n");
}

/********************************************************************/

// --remove-files: Remove all files given on the command line if their
// name (and possibly relative pathnames, if provided) exist in the
// --from path. What this means is that if one of the files provided on
// the command line is, e.g. frankie/picture.jpg and --from is /people,
// and there exists a file /people/frankie/picture.jpg in the NFI, it will
// be removed.
void cmdRemoveFiles (NFIFile *nfi , CommandStruct *command , ConfigInfo *config)
{
    // print out what we're going to be doing
    printRemoveCommand (command , config);

    // Loop over all files that were given on the command line. For each one, construct
    // a full path inside the NFI archive, and then try to remove the file from it.
    for (int i = 0 ; i < config->cmdlineFileCount ; i++)
    {
        char thisFile[NFI_NAMELEN + 1];

        // Build a full filename, which consists of the from path, and the path given
        // on the command line (which may also contain path information, which will
        // be considered relative to fromPath)
        sprintf (thisFile , "%s%s" , command->fromPath , config->cmdlineFiles[i]);

        // Try to remove the file now. If it worked, the file changed.
        if (nfi->pathStruct->removeFile (thisFile))
            nfi->changeCount++;
    }

    // All done. If we are outputting what we're doing, seperate this command from any that might follow
    if (config->isTesting)
        printf ("\n");
}

/********************************************************************/

