#ifndef __CONFIGdotH__
#define __CONFIGdotH__

/********************************************************************/

// This enumeration indicates if we should keep spaces in incoming files,
// convert them to underscores or remove them.
typedef enum
{
    SPACES_KEEP=0,
    SPACES_CONVERT,
    SPACES_REMOVE
} SpaceConversion;

// This enumeration indicates what sort order things should use in the
// listing mode. The default is natural, which means to leave things
// in the order they're stored in the file.
typedef enum
{
    SORT_NATURAL=0,
    SORT_NAME,
    SORT_SIZE,
    SORT_OFFSET,
    SORT_PATH
} SortStyle;

// This enumeration is used to provide unique values to different
// long command line options so that we can distinguish them.
typedef enum
{
    CMD_FILE=1000, 
    CMD_CREATE,
    CMD_UPDATE,
    CMD_LIST,
    CMD_VERBOSE,
    CMD_TESTING,
    CMD_VERSION,
    CMD_HELP,
    CMD_TMPDIR,
    CMD_KEEPCASE,
    CMD_LOWCASE,
    CMD_KEEPSPACE,
    CMD_CONVSPACE,
    CMD_DELSPACE,
    CMD_SORT,
    CMD_DESCEND,
    CMD_BRIEF,
    CMD_NOSIZE,
    CMD_HUMAN,
    CMD_FROM,
    CMD_TO,
    CMD_THEN,
    CMD_CLOBBER,
    CMD_NOCLOBBER,
    CMD_DEBUG,

    CMD_ADD,
    CMD_ADDRECURSE,
    CMD_ADDFILES,
    CMD_ADDDIR,
    CMD_ADDDIRRECURSE,

    CMD_REMOVE,
    CMD_REMOVERECURSE,
    CMD_REMOVEFILES,
    CMD_REMOVEDIR,
    CMD_REMOVEDIRRECURSE,

    CMD_MOVE,
    CMD_MOVERECURSE,
    CMD_MOVEFILES,
    CMD_MOVEDIR,
    CMD_MOVEDIRRECURSE,

    CMD_EXTRACTFILES,
    CMD_EXTRACTDIR,
    CMD_EXTRACTDIRRECURSE,
    CMD_EXTRACT,
    CMD_EXTRACTRECURSE,
} CommandType;

// An instance of this structure represents a command that was
// parsed from the command line. This contains a copy of what
// all of the relevant global options were at the time the command
// was concluded. This allows the user to change options on a
// per-command basis.
typedef struct
{
    // This command and it's argument
    CommandType cmd;
    char *arg;

    // The to and from paths
    char *fromPath;
    char *toPath;

    // if haveRegex is true, then compiledRegex is a compiled
    // regex structure taken from the arg of the command. Otherwise,
    // compiledRegex should be ignored.
    SDL_bool haveRegex;
    regex_t compiledRegex;

    // Should we clobber files that exist in the archive with files
    // being added?
    SDL_bool clobberIncoming;

    // Should we keep the case of added files? (false means to convert
    // to lowercase)
    SDL_bool keepIncomingCase;

    // What action to take with spaces on incoming additions.
    SpaceConversion spaceAction;
} CommandStruct; 

// An instance of this structure is allocated and returned by the code
// that parsed the command line arguments. This contains all of the
// information needed to run the program.
typedef struct
{
    // The filename of the NFI file that we should be working on.
    char *nfiFilename;

    // The location for temporary files
    char *tmpDir;

    // Locations of from and to paths
    char *fromPath;
    char *toPath;

    // Are we creating, updating or listing?
    SDL_bool isCreating;
    SDL_bool isUpdating;
    SDL_bool isListing;

    // Should we outpu extra data (for debugging?)
    SDL_bool isDebugging;

    // Should we be verbose?
    SDL_bool isVerbose;

    // Should we do anything?
    SDL_bool isTesting;

    // Should we show version information or help?
    SDL_bool showVersion;
    SDL_bool showHelp;

    // Should we keep the case of added files? (false means to convert
    // to lowercase)
    SDL_bool keepIncomingCase;

    // What action to take with spaces on incoming additions.
    SpaceConversion spaceAction;

    // Should we clobber files that exist in the archive with files
    // being added?
    SDL_bool clobberIncoming;

    // What type of sort do we want?
    SortStyle sortType;
    SDL_bool isDescendingSort;

    // Display brief listings? with or without a size?
    SDL_bool useBriefListing;
    SDL_bool omitSize;

    // Display human readable sizes?
    SDL_bool useHumanSizes;

    // When parsing the command line these list the last command that
    // was seen, and it's argument (or NULL if it took no argument).
    // These are all the add/remove/move/extract commands.
    CommandType currentCommand;
    char *currentCommandArg;

    // A dynamic array of commands. The array is commandTableSize
    // entries large and stores commandSize entries. If we need to
    // add an entry and the table isn't big enough, we allocate a
    // bigger one.
    int commandCount;
    int commandTableSize;
    CommandStruct *commands;

    // These variables are used to store files that are provided 
    // on the command line outside of all options. Some commands
    // work on filenames of this sort.
    int cmdlineFileCount;
    char **cmdlineFiles;
} ConfigInfo; 

/********************************************************************/

// Prints out help on how to use the tool to stdout.
void printCommandHelp (void);

// Convert an integer that represnets a command into a string suitable
// for display purposes.
//
// If value is one of the CMD_* values, the appropriate long option is
// returned. If it is a character constant, then the appropriate short
// option is returned.
char *optionNameFromValue (int value);

// Takes the command line parameters given to us and returns back a 
// configuration structure, or NULL on error, in which case an error will
// be printed.
//
// The config information is not fully parsed if a request for version
// information or help is given. If this returns success, you should check
// for one of those situations before continuing on.
ConfigInfo *openConfigInfo (int argc, char **argv);

// Close up and release all memory used by the config information
// passed in. This works even if info is NULL or only partially
// filled out
void closeConfigInfo (ConfigInfo *info);

/********************************************************************/

#endif

