#ifndef __FILESCANNERdotH__
#define __FILESCANNERdotH__

/********************************************************************/

// These define how many entries initially are allocated
// in a ScannedFiles structure, and how many they grow
// by when we run out of space. Since we only store pointers, we
// set these fairly high, since if there are a lot of files or a lot
// of directories to scan, there could be a lot of calls to realloc
// to resize the list, which kills performance.
#define NAME_GROW_SIZE 10000
#define DIR_GROW_SIZE 10000

/********************************************************************/

// This class does the work of scanning a path for all files contained
// and storing their names. Optionally, you can provide a regex which
// a filename (ONLY) must match in order to be added to the list of scanned
// files. You may also optionally specify that all paths under the current
// path be searched as well.
//
// The search works in a non-recursive way so that even on very deep
// directory structures we don't run into problems with having too many
// directory handles open. Because of this, the files that are found may
// not all be in a "sensible" file system defined order.
//
// The files searched contain the name of the file and all paths relative
// to the path that was initially searched. So that if you search "../",
// you get the names of all files and subdirectories, but they do not
// start with "../".
//
// This will only scan regular files and ignores symbolic links instead
// of following them, since this can lead the scanner into an infinit search.
class ScannedFiles
{
    public:
        // Scan the provided path. If regex is non-null, use it to filter what files
        // get added. if verbose is true, announce directories as they are scanned.
        // If recurse is true, scan all subdirectories of the given path and add files
        // from there as well.
        ScannedFiles (const char *path, regex_t *regex, SDL_bool verbose, SDL_bool recurse);

        // Take a list of pre-scanned files, relative to the current directory. Each name
        // will be checked to see if it exists or not. If it is, it will be added. Any
        // of the files specified that are paths will be recursed into.
        ScannedFiles (const char **files, int count, SDL_bool verbose);

        ~ScannedFiles ();

        // The directory that was scanned for files
        const char *scannedDir (void) { return rootDir; }

        // Lists the total number of matches found for files, and allows you to
        // access them. If a regex was specified, the found files only contain
        // filenames that matched the regex.
        int filesFound (void) { return nameSize; }
        const char *fileAtIndex (int i) { return names[i]; }

        // Lists the total number of directories found and allows you to
        // access them. Note that this is the number of directories seen,
        // even if some directories were ignored because no files were
        // in them, or no files that matched a regex were in them.
        int dirsFound (void) { return dirSize; }
        const char *dirAtIndex (int i) { return dirs[i]; }

    private:
        // Does the actual work of finding files in a specific path
        void scanPath (const char *path);

        // Scan files for inclusion in the list. If initialScan is TRUE, then
        // we scan the rootDir given in the constructor and add all files and
        // paths specified there. If initialScan is FALSE, this step is skipped
        // and it is assumed that the internal lists have been filled out with
        // some other information to begin with.
        //
        // If recurse is TRUE, all directories found in the initial scan (or placed
        // in the directory list by code prior to this call) will be scanned the
        // same as the rootDir was above.
        void ScannedFiles::findFilesInPath (SDL_bool initialScan , SDL_bool recurse);

        // Does a stat to determine if text is a file or a directory
        // and adds accordingly. This will not add special files.
        void addText (const char *path , const char *text);

        // These internally grow the size of lists.
        void growNameList (void);
        void growDirList (void);

        // A dynamic array used to store names. maxNameSize
        // is the maximum number of names that can be stored
        // before we need to grow the table, and nameSize is
        // how many names are being stored.
        int maxNameSize;
        int nameSize;
        char **names;

        // The same as above, but this stores directory names
        // instead of filenames.
        int maxDirSize;
        int dirSize;
        char **dirs;

        // The directory all files are relative to
        char *rootDir;
        int rootDirLen;

        // If non-null, use the compiled regex expression to match on files
        // being added.
        regex_t *regex;

        // Should we be verbose while scanning?
        SDL_bool bVerbose;
};

/********************************************************************/

#endif

