#ifndef __NFIPATHdotH__
#define __NFIPATHdotH__

/********************************************************************/

// Define the number of bytes in the extraction buffer.
#define EXTRACT_BUF_BYTES 65535

/********************************************************************/

// This class represents the directory structure inside an NFI file
// in a way that it is easy to scan around for files and directories
// and act on them accordingly. We store the paths of the file in a
// hierarchal way so that it's easier to locate all files that we want.
// All files stored maintain a pointer to their NFI information in
// the loaded NFI file, so that access to that information is easy.
//
// When files are added, they must provide a pointer to their
// NFI info. When files are removed, they remain in the path they
// were in, but their NFI info flags them as removed for other
// operations. When move operations occur, the appropriate file tree
// entries are physically moved to the proper location, both here
// and inside the NFI itself.
class NFIPathStructure
{
    public:
        // Construct a new structure based off of a preloaded file.
        // This also flags things as verbose or not. This affects
        // what the add/remove/move functions display.
        NFIPathStructure (NFIFile *file , SDL_bool verbose);
        ~NFIPathStructure ();

        // Output the contents of the NFI structure to stdout
        void displayContents ();

        // Add a file at the path given in the node. When clobber is TRUE, 
        // if there is a file with the name given at the path given, it will 
        // be replaced with the new file (that is, the NFI info will be copied 
        // overtop of the old file) Otherwise, such a file will not be touched.
        //
        // When a file is clobbered, the node it was stored at is re-used
        // so that no new links need to be made. The info in the node will
        // be overwritten by the info passed here. However, the info for
        // the file being clobbered will have it's type set to removed.
        //
        // The functions return true if the add succeded and false if not
        // (that is, if the file addition would have clobbered something 
        // and clobber was false).
        SDL_bool addFile (NFIParsedDirEntry *info , SDL_bool clobber = SDL_FALSE);

        // Flag files at the given path as removed. If recursive is true, also
        // flag files in every path underneath this one. If a compiled regex is specified (non-null)
        // the filenames in the given path (or paths) must match the regex to be removed. otherwise,
        // they are removed unconditionally.
        //
        // The second version takes a fully qualified internal name and tries to remove it.
        //
        // The return values are TRUE if at least one file was moved, and FALSE if not.
        SDL_bool removeFiles (const char *path , regex_t *regex , SDL_bool recursive);
        SDL_bool removeFile (const char *name);

        // Move all files from the given fromPath node to the toPath node. If recursive is
        // TRUE, then directory nodes are also moved.
        //
        // All affected files have their NFI info updated to reflect their new path.
        //
        // If clobber is false, any file in fromPath that already exists in toPath 
        // will not be moved (that is, it's NFI info won't be updated and the path
        // it was stored in won't be fully moved). Otherwise, everything is moved
        // wholesale to the toPath.
        //
        // The move will skip files marked as removed. If a file is marked as added, then it's internal
        // path and location in the path tree will be altered, but the file itself will remain flagged
        // as added. Otherwise, the file will be flagged as moved. 
        //
        // If a regex is given, then a file will only be considered for moving if it's name
        // matches the regex. If regex is NULL, all found files are considered for the move.
        //
        // The second version takes a fully qualified filename and a path into the NFI, and
        // tries to move only that file.
        //
        // Return value tells you if at least one file was moved or not.
        SDL_bool moveFiles (const char *fromPath , const char *toPath , regex_t *regex , SDL_bool recursive, SDL_bool clobber);
        SDL_bool moveFile (const char *name , const char *toPath , SDL_bool clobber);

        // Extract the file or files at fromPath in the NFI archive to toPath in the local filesystem. Files stored
        // exactly in fromPath will be stored directly in toPath, and subdirectories of fromPath will be stored in
        // subdirectories of toPath. 
        //
        // If a regex is given, the filename of a file must match it to be extracted. If the regex is NULL, all files
        // found will be extracted.
        //
        // recursive identifies if all subdirectories of fromPath should be extracted, or just the files stored in fromPath.
        //
        // clobber specifies if it is ok for an extraction of a file to overwrite a file on the local filesystem when it
        // already exists. If a file exists and clobber is false, it will not be extracted.
        //
        // A file will only be considered for extraction if it is untouched, moved or added. A file flagged for removal
        // will not be considered for extraction. A file that is flagged for addition will be copied from the source
        // location to the destination location as if it already existed in the archive.
        //
        // The second version takes a fully qualified filename and a path to the local filesystem. The file will be
        // extracted if it can be found.
        //
        // Neither of these functions return a value, because wether a file was extracted or not does not alter the archive
        // at all.
        void extractFiles (const char *fromPath , const char *toPath , regex_t *regex , SDL_bool recursive , SDL_bool clobber);
        void extractFile (const char *name , const char *toPath , SDL_bool clobber);

    private:
        // Scan the provided NFIFile structure and add all files and paths in it to ourselves.
        void scanFile (NFIFile *file);
       
        // Find and return a pointer to a given path. If the path
        // given doesn't exist yet in the path structure, it will
        // be created for you. If the path you pass in is "/",
        // the return value is NULL. The methods below that take
        // a PathEntry know what this special value means.
        PathEntry *findPath (const char *path);

        // Return a full path from root to the passed in node in a static
        // char buffer. This makes it quite, quite unthreadsafe. If the node
        // passed in is NULL, the string "/" is returned.
        const char *pathFromNode (PathEntry *node);

        // Add a new node to the given node. if isFile is TRUE, the new node is added to the file
        // list of the provided node, otherwise it is added to the directory list of the added
        // node. In either case, the newly inserted node is returned back to you. The new node will
        // be fully initialized, though if it is a file node it is up to the caller to add in the
        // NFI info.
        PathEntry *addNodeToPath (PathEntry *node , const char *name , SDL_bool isFile);

        // Search the given node to see if it contains the name provided. If it
        // does, return a pointer to the node containing that name. Otherwise,
        // return NULL. NULL is returned if name is "/" or "" as well, since that
        // is technically the node that houses the root dir.
        //
        // If isFile is true, the file list in the passed in node is searched. 
        // Otherwise, the directory list is searched.
        //
        // Note that the passed in node can be NULL (root or /).
        PathEntry *findEntryNamed (PathEntry *node , const char *name , SDL_bool isFile);

        // Given a path entry (which may be NULL for the root of the path system), return back a pointer to the
        // file or directory list of that node. These helpers exist because the data for the root appears in
        // members of the class, but the data for other paths appear in the nodes themselves.
        PathEntry *getFileList (PathEntry *path);
        PathEntry *getDirectoryList (PathEntry *path);

        // Display a node in the tree.
        void printNode (PathEntry *node , int indent);

        // Takes a path node that is the head of a list of files, and removes every file in the list.
        void removeFileList (PathEntry *head);

        // Takes a path node that is the head of a list of directories. All directories and subdirectories
        // of the given node are removed, as well as any files they are storing.
        void removeDirList (PathEntry *head);

        // Rebuild the file list in the passed in node (which may be NULL for root) so that any 
        // nodes that it is storing that are flagged for deletion will be removed.
        void cleanFileList (PathEntry *path);

        // Check to see if there is a file at the path provided. If there is, return TRUE. Otherwise,
        // return false. This checks the local filesystem. The name passed can therefore be absolute,
        // or relative to the current directory, as needed.
        SDL_bool realFileExists (const char *name);

        // Take a path that is either relative to the current directory or
        // an absolute path. Make sure that all paths along this path
        // exist. The name passed in may optionally contain a filename spec.
        // If it doesn't, the name must terminate in a '/' character.
        void createFullPath (const char *name);

        // Internal functions for manipulating the archive.
        SDL_bool addFile (PathEntry *path , char *filename , NFIParsedDirEntry *info , SDL_bool clobber);
        SDL_bool removeFiles (PathEntry *path , regex_t *regex , SDL_bool recursive);
        SDL_bool moveFiles (PathEntry *fromPath , PathEntry *toPath , regex_t *regex , SDL_bool recursive, SDL_bool clobber);
        SDL_bool performMove (PathEntry *fromPath , PathEntry *toPath , PathEntry *fileNode , regex_t *regex , SDL_bool clobber);
        void extractFiles (PathEntry *fromPath , const char *toPath , regex_t *regex , SDL_bool recursive, SDL_bool clobber);
        SDL_bool performExtract (PathEntry *fileNode , const char *toPath , regex_t *regex , SDL_bool clobber);

        // Should we be verbose?
        SDL_bool bVerbose;

        // The class represents the root (/) directory in the archive because it
        // would be silly to use a node for that and have to index it every time
        // anything happened. So these store all directories off the root,and all 
        // of the files stored directly in the root.
        PathEntry *directories;
        PathEntry *files;

        // The archive that we represent. We stash this when we scan it for it's
        // structure in case we need it for extraction operations.
        NFIFile *archive;

        // This buffer starts out NULL. The first time a file is extracted, it is
        // allocated to be a buffer of size EXTRACT_BUF_BYTES. This defers
        // allocation until/when we need it.
        Uint8 *extractBuffer;
};

/********************************************************************/

// Take a mutable path string which may or may not begin with a '/' character. Find
// the next path component and return it. startPos is the position in the string to
// start scanning. The initial call should call this with 0. On successful return,
// it will be set so that successive calls to nextPathPart can be made using the
// value it stores.
//
// Note that the input string will be fiddled with, so it must be an editable
// copy of the string that you will use for no other purpose.
//
// Returns a pointer to a path component on success or NULL if the path is done now.
char *nextPathPart (char *path , int *startPos);

// Copy dataLength bytes of data from fromHandle to toHandle, using the buffer of the size
// given. The two bools given will be set to FALSE on entry. On exit, if either one is TRUE, it
// tells you what kind of error occured. If both are false, there was no error.
void copyFileData (FILE *fromHandle , FILE *toHandle , Uint32 dataLength , Uint8 *buffer , Uint32 bufferLen , SDL_bool *readError , SDL_bool *writeError);

/********************************************************************/

#endif

