#ifndef __NFISPECdotH__
#define __NFISPECdotH__

/********************************************************************/

// These data types are specific to the NFI Handling code. The sizes
// were mirrored from the NFI Tool code that was most current at the
// time that it was written. That means that this code can only read
// older files until someone specifically changes it (assuming the NFI
// format ever changes).
#define NFI_NAMELEN    260
#define NFI_VERSION      1
#define NFI_HEADERID "NFI"

/********************************************************************/

// This enumeration is used by us to determine what (if anything)
// has happened to a NFIParsedDirEntry. This tells us if the entry
// is untouched, or contains data that we need to take into account
// when we are saving out the file (for example, that it represents
// a newly added file that needs to be copied from somewhere).
typedef enum
{
    // This entry has not been modified in any way
    NFI_UNTOUCHED=0,

    // The entry represnets a file being added to the file
    NFI_ADDED_FILE,

    // The entry represents a file that was removed from the file
    NFI_REMOVED_FILE,

    // The entry represents a file that has had it's internal path
    // changed in some way.
    NFI_MOVED_FILE,
} NFIEntryType;

// The definition of an NFI File header
typedef struct
{
    char  id[4];
    char  version;
    Uint32 reserved;
    Uint32 dirOffs;
} NFIFileHeader;

// The definition of an NFI directory header
typedef struct
{
    Uint32 dirSize;
    Uint32 reserved;
} NFIDirHeader;

// The definition of an NFI directory entry
// as it is stored in the file.
typedef struct NFIDirEntry
{
    char  name[NFI_NAMELEN];
    Uint32 offset;
    Uint32 length;
    char  reserved[8];
} NFIDirEntry;

// The definition of an NFI directory entry
// after it has been loaded. We split the name
// into a name and path field, plus add some extra
// data to it.
typedef struct
{
    // These are the same as NFIDirEntry, except that
    // the path has been split out of the name
    char  name[NFI_NAMELEN];
    char  path[NFI_NAMELEN];
    Uint32 offset;
    Uint32 length;
    char  reserved[8];

    // The type of the entry. Default is untouched but
    // if we move/remove/add a file, that will be reflected
    // here. When we create a directory entry to add in a new
    // file, we set the type to "add" and then store the name
    // of the file that is to be added in sourceName so that we
    // know where to go for the data at write time.
    NFIEntryType type;
    char sourceName[NFI_NAMELEN];

    // This maintains a pointer to the PathEntry (inside the current
    // instance of NFIPathStructure) that is representing this file.
    // When an add operation occurs, we realloc the list of these
    // Parsed entries to be big enough to hold the added files. At
    // that time, the pointers that PathEntry structures hold become
    // invalid. This field allows us to "tag" what entries are
    // reffering to what, so that after a realloc of the global
    // table, we can fix their info pointers.
    struct PathEntryTAG         *entryPtr;
} NFIParsedDirEntry;

// This is used to store NFI path information in the NFIPathStructure
// class. This can store either information about a file in an NFI 
// file, or a path in an NFI file. We use these to contruct a logical
// tree of files and paths, so that operations on NFI files are
// faster.
//
// The lists these are used in are doubly linked lists for ease in
// navigation and operations that involve moving nodes around in 
// the tree. If a tree structure of some sort was used, operations
// like remove and move would be faster in larger files.
typedef struct PathEntryTAG
{
    // The name of this path entry. For paths, this is the name of a 
    // path segment (e.g., the "gfx" in the full path "/gfx/icons").
    // For a name, it's the name of the file being stored.
    char *name;

    // The payload of this node. If the node is storing a path, then:
    //   o files is a sublist of all files stored at this path (NULL == none)
    //   o directories is a sublist of all directories that are under this path (NULL = none)
    //
    // If node is storing a filename then:
    //   o info is an NFI info struct for the contained file. This is actually
    //     stored in the NFI directory but aliased here so that it's easy to get
    //     at for various operations.
    union 
    {
        struct 
        {
            PathEntryTAG *files;
            PathEntryTAG *directories;
        };

        NFIParsedDirEntry *info;
    };

    // This flag is used to flag nodes in various tree structures as nodes
    // which should be removed after some operations have completed. During
    // these operations, we are traversing the tree and do not want to remove
    // nodes out from under ourselves. So instead they are flagged and cleaned
    // up later on. This is initialized to false whenever a new PathEntry is
    // created and set to true when we want a node to be deleted.
    SDL_bool shouldDelete;

    // This points to the path entry in the list that is the parent of
    // the file or path that is being stored. (e.g., if we are storing
    // the "icons" part of "/gfx/icons", this points to the entry that
    // is storing the "gfx" part).
    //
    // If this is NULL, then the directory or file in question is being 
    // stored in the root (/) directory.
    PathEntryTAG *parent;

    // These link to the next and previous nodes in the list of files
    // or paths.
    PathEntryTAG *next;
    PathEntryTAG *prev;
} PathEntry;

// Forward declare these because they come from a header file that
// needs us to exist, and the circle must end somewhere.
class NFIPathStructure;

// The definition of an actual NFI File. Contains a file header, a
// directory header, and a dynamically sized array that contains the
// actual directory of the file. 
//
// Whenever a new node is added to an NFIPathStructure tree, and it
// gets a pointer to a struct like this, we set entryPtr to be a
// pointer to the PathEntry that owns us, so that if we reallocate
// the directory of the NFI file, we can fix things.
typedef struct 
{
    char              name[NFI_NAMELEN];
    NFIFileHeader     fileHeader;
    NFIDirHeader      dirHeader;
    NFIParsedDirEntry *directory;
    FILE              *fileHandle;
    Uint32            longestFilename;
    Uint32            longestPathComponent;
    SDL_bool          newFile;
    SDL_bool          movesOnly;
    Uint32            changeCount;
    NFIPathStructure  *pathStruct;
} NFIFile;

/********************************************************************/

#endif

